"""
Tests that each element in an anndata is written correctly
"""

from __future__ import annotations

import re
from contextlib import ExitStack, nullcontext
from importlib.metadata import version
from pathlib import Path
from typing import TYPE_CHECKING

import h5py
import numpy as np
import pandas as pd
import pytest
import zarr
from packaging.version import Version
from scipy import sparse

import anndata as ad
from anndata._io.specs import _REGISTRY, IOSpec, get_spec
from anndata._io.specs.registry import IORegistryError
from anndata._io.zarr import open_write_group
from anndata.compat import CSArray, CSMatrix, ZarrGroup, _read_attr, is_zarr_v2
from anndata.experimental import read_elem_lazy
from anndata.io import read_elem, write_elem
from anndata.tests.helpers import (
    GEN_ADATA_NO_XARRAY_ARGS,
    as_cupy,
    as_cupy_sparse_dask_array,
    as_dense_cupy_dask_array,
    assert_equal,
    check_all_sharded,
    gen_adata,
    visititems_zarr,
)

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path
    from typing import Literal, TypeVar

    from anndata._types import GroupStorageType
    from anndata.compat import H5Group

    G = TypeVar("G", H5Group, ZarrGroup)


PANDAS_3 = Version(version("pandas")) >= Version("3rc0")


@pytest.fixture
def exit_stack() -> Generator[ExitStack, None, None]:
    with ExitStack() as stack:
        yield stack


@pytest.fixture
def store(diskfmt, tmp_path) -> H5Group | ZarrGroup:
    if diskfmt == "h5ad":
        file = h5py.File(tmp_path / "test.h5ad", "w")
        store = file["/"]
    elif diskfmt == "zarr":
        store = open_write_group(tmp_path / "test.zarr")
    else:
        pytest.fail(f"Unknown store type: {diskfmt}")

    try:
        yield store
    finally:
        if diskfmt == "h5ad":
            file.close()


sparse_formats = ["csr", "csc"]
SIZE = 50
DEFAULT_SHAPE = (SIZE, SIZE * 2)


@pytest.fixture(params=sparse_formats)
def sparse_format(request: pytest.FixtureRequest) -> Literal["csr", "csc"]:
    return request.param


def create_dense_store(
    store: str, *, shape: tuple[int, ...] = DEFAULT_SHAPE
) -> H5Group | ZarrGroup:
    X = np.random.randn(*shape)

    write_elem(store, "X", X)
    return store


def create_sparse_store(
    sparse_format: Literal["csc", "csr"], store: G, shape=DEFAULT_SHAPE
) -> G:
    """Returns a store

    Parameters
    ----------
    sparse_format
    store

    Returns
    -------
        A store with a key, `X` that is simply a sparse matrix, and `X_dask` where that same array is wrapped by dask
    """
    import dask.array as da

    X = sparse.random(
        shape[0],
        shape[1],
        format=sparse_format,
        density=0.01,
        random_state=np.random.default_rng(),
    )
    X_dask = da.from_array(
        X,
        chunks=(100 if format == "csr" else SIZE, SIZE * 2 if format == "csr" else 100),
    )

    write_elem(store, "X", X)
    write_elem(store, "X_dask", X_dask)
    return store


@pytest.mark.parametrize(
    ("value", "encoding_type"),
    [
        pytest.param(None, "null", id="none"),
        pytest.param("hello world", "string", id="py_str"),
        pytest.param(np.str_("hello world"), "string", id="np_str"),
        pytest.param(np.array([1, 2, 3]), "array", id="np_arr_int"),
        pytest.param(
            np.array(["hello", "world"], dtype=object), "string-array", id="np_arr_str"
        ),
        pytest.param(1, "numeric-scalar", id="py_int"),
        pytest.param(True, "numeric-scalar", id="py_bool"),
        pytest.param(1.0, "numeric-scalar", id="py_float"),
        pytest.param({"a": 1}, "dict", id="py_dict"),
        pytest.param(
            lambda: gen_adata((3, 2), **GEN_ADATA_NO_XARRAY_ARGS),
            "anndata",
            id="anndata",
        ),
        pytest.param(
            sparse.random(5, 3, format="csr", density=0.5),
            "csr_matrix",
            id="sp_mat_csr",
        ),
        pytest.param(
            sparse.random(5, 3, format="csc", density=0.5),
            "csc_matrix",
            id="sp_mat_csc",
        ),
        pytest.param(pd.DataFrame({"a": [1, 2, 3]}), "dataframe", id="pd_df"),
        pytest.param(
            pd.Categorical([*"aabccedd", pd.NA]),
            "categorical",
            id="pd_cat_np_str",
        ),
        pytest.param(
            pd.Categorical(list("aabccedd"), ordered=True),
            "categorical",
            id="pd_cat_np_str_ord",
        ),
        pytest.param(
            pd.array([*"aabccedd", pd.NA], dtype="string").astype("category"),
            "categorical",
            id="pd_cat_pd_str",
        ),
        pytest.param(
            pd.Categorical([1, 2, 1, 3], ordered=True), "categorical", id="pd_cat_num"
        ),
        pytest.param(
            pd.array(["hello", "world"], dtype="string"),
            "nullable-string-array",
            id="pd_arr_str",
        ),
        pytest.param(
            pd.array(["hello", "world", pd.NA], dtype="string"),
            "nullable-string-array",
            id="pd_arr_str_mask",
        ),
        pytest.param(
            pd.arrays.IntegerArray(
                np.ones(5, dtype=int), mask=np.array([True, False, True, False, True])
            ),
            "nullable-integer",
            id="pd_arr_int_mask",
        ),
        pytest.param(pd.array([1, 2, 3]), "nullable-integer", id="pd_arr_int"),
        pytest.param(
            pd.arrays.BooleanArray(
                np.random.randint(0, 2, size=5, dtype=bool),
                mask=np.random.randint(0, 2, size=5, dtype=bool),
            ),
            "nullable-boolean",
            id="pd_arr_bool_mask",
        ),
        pytest.param(
            pd.array([True, False, True, True]), "nullable-boolean", id="pd_arr_bool"
        ),
        pytest.param(
            zarr.ones((100, 100), chunks=(10, 10)),
            "array",
            id="zarr_dense_array",
        ),
        pytest.param(
            create_dense_store(
                h5py.File("test1.h5", mode="w", driver="core", backing_store=False)
            )["X"],
            "array",
            id="h5_dense_array",
        ),
        # pytest.param(bytes, b"some bytes", "bytes", id="py_bytes"), # Does not work for zarr
        # TODO consider how specific encodings should be. Should we be fully describing the written type?
        # Currently the info we add is: "what you wouldn't be able to figure out yourself"
        # but that's not really a solid rule.
        # pytest.param(bool, True, "bool", id="py_bool"),
        # pytest.param(bool, np.bool_(False), "bool", id="np_bool"),
    ],
)
def test_io_spec(store: GroupStorageType, value, encoding_type) -> None:
    if callable(value):
        value = value()

    key = f"key_for_{encoding_type}"
    with ad.settings.override(allow_write_nullable_strings=True):
        write_elem(store, key, value, dataset_kwargs={})

    assert encoding_type == _read_attr(store[key].attrs, "encoding-type")

    from_disk = read_elem(store[key])
    assert_equal(value, from_disk)
    assert get_spec(store[key]) == _REGISTRY.get_spec(value)


@pytest.mark.parametrize(
    ("value", "encoding_type"),
    [
        pytest.param(np.asarray(1), "numeric-scalar", id="scalar_int"),
        pytest.param(np.asarray(1.0), "numeric-scalar", id="scalar_float"),
        pytest.param(np.asarray(True), "numeric-scalar", id="scalar_bool"),  # noqa: FBT003
        pytest.param(np.asarray("test"), "string", id="scalar_string"),
    ],
)
def test_io_spec_compressed_scalars(store: G, value: np.ndarray, encoding_type: str):
    key = f"key_for_{encoding_type}"
    write_elem(
        store, key, value, dataset_kwargs={"compression": "gzip", "compression_opts": 5}
    )

    assert encoding_type == _read_attr(store[key].attrs, "encoding-type")

    from_disk = read_elem(store[key])
    assert_equal(value, from_disk)


# Can't instantiate cupy types at the top level, so converting them within the test
@pytest.mark.gpu
@pytest.mark.parametrize(
    ("value", "encoding_type"),
    [
        (np.array([1, 2, 3]), "array"),
        (np.arange(12).reshape(4, 3), "array"),
        (sparse.random(5, 3, format="csr", density=0.5), "csr_matrix"),
        (sparse.random(5, 3, format="csc", density=0.5), "csc_matrix"),
    ],
)
@pytest.mark.parametrize("as_dask", [False, True])
def test_io_spec_cupy(store, value, encoding_type, as_dask):
    if as_dask:
        if isinstance(value, CSMatrix):
            value = as_cupy_sparse_dask_array(value, format=encoding_type[:3])
        else:
            value = as_dense_cupy_dask_array(value)
    else:
        value = as_cupy(value)

    key = f"key_for_{encoding_type}"
    write_elem(store, key, value, dataset_kwargs={})

    assert encoding_type == _read_attr(store[key].attrs, "encoding-type")

    from_disk = as_cupy(read_elem(store[key]))
    assert_equal(value, from_disk)
    assert get_spec(store[key]) == _REGISTRY.get_spec(value)


def test_dask_write_sparse(sparse_format, store):
    x_sparse_store = create_sparse_store(sparse_format, store)
    X_from_disk = read_elem(x_sparse_store["X"])
    X_dask_from_disk = read_elem(x_sparse_store["X_dask"])

    assert_equal(X_from_disk, X_dask_from_disk)
    assert_equal(dict(x_sparse_store["X"].attrs), dict(x_sparse_store["X_dask"].attrs))

    assert x_sparse_store["X_dask/indptr"].dtype == np.int64
    assert x_sparse_store["X_dask/indices"].dtype == np.int64


def test_read_lazy_2d_dask(sparse_format, store):
    arr_store = create_sparse_store(sparse_format, store)
    X_dask_from_disk = read_elem_lazy(arr_store["X"])
    X_from_disk = read_elem(arr_store["X"])

    assert_equal(X_from_disk, X_dask_from_disk)
    random_int_indices = np.random.randint(0, SIZE, (SIZE // 10,))
    random_int_indices.sort()
    index_slice = slice(0, SIZE // 10)
    for index in [random_int_indices, index_slice]:
        assert_equal(X_from_disk[index, :], X_dask_from_disk[index, :])
        assert_equal(X_from_disk[:, index], X_dask_from_disk[:, index])
    random_bool_mask = np.random.randn(SIZE) > 0
    assert_equal(
        X_from_disk[random_bool_mask, :], X_dask_from_disk[random_bool_mask, :]
    )
    random_bool_mask = np.random.randn(SIZE * 2) > 0
    assert_equal(
        X_from_disk[:, random_bool_mask], X_dask_from_disk[:, random_bool_mask]
    )

    assert arr_store["X_dask/indptr"].dtype == np.int64
    assert arr_store["X_dask/indices"].dtype == np.int64


@pytest.mark.parametrize(
    ("n_dims", "chunks"),
    [
        (1, (10,)),
        (1, (40,)),
        (2, (10, 10)),
        (2, (40, 40)),
        (2, (20, 40)),
        (1, None),
        (2, None),
        (2, (40, -1)),
        (2, (40, None)),
    ],
)
def test_read_lazy_subsets_nd_dask(store, n_dims, chunks):
    arr_store = create_dense_store(store, shape=DEFAULT_SHAPE[:n_dims])
    X_dask_from_disk = read_elem_lazy(arr_store["X"], chunks=chunks)
    X_from_disk = read_elem(arr_store["X"])
    assert_equal(X_from_disk, X_dask_from_disk)

    random_int_indices = np.random.randint(0, SIZE, (SIZE // 10,))
    random_int_indices.sort()
    random_bool_mask = np.random.randn(SIZE) > 0
    index_slice = slice(0, SIZE // 10)
    for index in [random_int_indices, index_slice, random_bool_mask]:
        assert_equal(X_from_disk[index], X_dask_from_disk[index])


@pytest.mark.xdist_group("dask")
@pytest.mark.dask_distributed
def test_read_lazy_h5_cluster(
    sparse_format: Literal["csr", "csc"],
    tmp_path: Path,
    local_cluster_addr: str,
) -> None:
    import dask.distributed as dd

    with h5py.File(tmp_path / "test.h5", "w") as file:
        store = file["/"]
        arr_store = create_sparse_store(sparse_format, store)
        X_dask_from_disk = read_elem_lazy(arr_store["X"])
        X_from_disk = read_elem(arr_store["X"])
    with dd.Client(local_cluster_addr):
        assert_equal(X_from_disk, X_dask_from_disk)


def test_undersized_shape_to_default(store: H5Group | ZarrGroup):
    shape = (1000, 50)
    arr_store = create_dense_store(store, shape=shape)
    X_dask_from_disk = read_elem_lazy(arr_store["X"])
    assert all(c <= s for c, s in zip(X_dask_from_disk.chunksize, shape, strict=True))
    assert X_dask_from_disk.shape == shape


@pytest.mark.parametrize(
    ("arr_type", "chunks", "expected_chunksize"),
    [
        pytest.param("dense", (10, 10), (10, 10), id="dense"),
        pytest.param("csc", (SIZE, 1), (SIZE, 1), id="singleton-minor"),
        pytest.param(
            "csr",
            (1, SIZE * 2),
            (1, SIZE * 2),
            id="singleton-major",
        ),
        pytest.param("csc", None, (SIZE, 100), id="csc-default"),
        pytest.param("csr", None, DEFAULT_SHAPE, id="csr-default"),
        pytest.param(
            "csr",
            (10, -1),
            (10, SIZE * 2),
            id="minus_one_minor",
        ),
        pytest.param("csc", (-1, 10), (SIZE, 10), id="minus_one_major"),
        pytest.param("csr", (10, None), (10, SIZE * 2), id="none_minor"),
        pytest.param("csc", (None, 10), (SIZE, 10), id="none_major"),
        pytest.param("csc", (None, None), DEFAULT_SHAPE, id="csc-none_all"),
        pytest.param("csr", (None, None), DEFAULT_SHAPE, id="csr-none_all"),
        pytest.param("csr", (-1, -1), DEFAULT_SHAPE, id="csr-minus_one_all"),
        pytest.param("csc", (-1, -1), DEFAULT_SHAPE, id="csc-minus_one_all"),
    ],
)
def test_read_lazy_2d_chunk_kwargs(
    store: H5Group | ZarrGroup,
    arr_type: Literal["csr", "csc", "dense"],
    chunks: None | tuple[int | None, int | None],
    expected_chunksize: tuple[int, int],
):
    if arr_type == "dense":
        arr_store = create_dense_store(store)
        X_dask_from_disk = read_elem_lazy(arr_store["X"], chunks=chunks)
    else:
        arr_store = create_sparse_store(arr_type, store)
        X_dask_from_disk = read_elem_lazy(arr_store["X"], chunks=chunks)
    assert X_dask_from_disk.chunksize == expected_chunksize
    X_from_disk = read_elem(arr_store["X"])
    assert_equal(X_from_disk, X_dask_from_disk)


def test_read_lazy_bad_chunk_kwargs(tmp_path):
    arr_type = "csr"
    with h5py.File(tmp_path / "test.h5", "w") as file:
        store = file["/"]
        arr_store = create_sparse_store(arr_type, store)
        with pytest.raises(
            ValueError, match=r"`chunks` must be a tuple of two integers"
        ):
            read_elem_lazy(arr_store["X"], chunks=(SIZE,))
        with pytest.raises(ValueError, match=r"Only the major axis can be chunked"):
            read_elem_lazy(arr_store["X"], chunks=(SIZE, 10))


@pytest.mark.parametrize("sparse_format", ["csr", "csc"])
def test_write_indptr_dtype_override(store, sparse_format):
    X = sparse.random(
        100,
        100,
        format=sparse_format,
        density=0.1,
        random_state=np.random.default_rng(),
    )

    write_elem(store, "X", X, dataset_kwargs=dict(indptr_dtype="int64"))

    assert store["X/indptr"].dtype == np.int64
    assert X.indptr.dtype == np.int32
    np.testing.assert_array_equal(store["X/indptr"][...], X.indptr)


@pytest.mark.parametrize(
    ("num_minor_axis", "expected_dtype"),
    [
        pytest.param(1, np.dtype("uint8"), id="one_col-expected_uint8_on_disk"),
        pytest.param(
            np.iinfo(np.uint8).max,
            np.dtype("uint8"),
            id="max_np.uint8-matching_dtype_on_disk",
        ),
        pytest.param(
            np.iinfo(np.int8).max,
            np.dtype("uint8"),
            id="max_np.int8-uint8_on_disk",
        ),
        pytest.param(
            np.iinfo(np.uint16).max,
            np.dtype("uint16"),
            id="max_np.uint16-matching_dtype_on_disk",
        ),
        pytest.param(
            np.iinfo(np.int16).max,
            np.dtype("uint16"),
            id="max_np.int16-uint16_on_disk",
        ),
        pytest.param(
            np.iinfo(np.uint32).max,
            np.dtype("uint32"),
            id="max_np.uint32-matching_dtype_on_disk",
        ),
        pytest.param(
            np.iinfo(np.int32).max,
            np.dtype("uint32"),
            id="max_np.int32-uint32_on_disk",
        ),
        pytest.param(
            np.iinfo(np.uint8).max + 1,
            np.dtype("uint16"),
            id="max_np.uint8_plus_one_cols-expected_uint16_on_disk",
        ),
        pytest.param(
            np.iinfo(np.uint16).max + 1,
            np.dtype("uint32"),
            id="max_np.uint16_plus_one_cols-expected_uint32_on_disk",
        ),
        pytest.param(
            np.iinfo(np.uint32).max + 1,
            np.dtype("uint64"),
            id="max_np.uint32_plus_one_cols-expected_uint64_on_disk",
        ),
        pytest.param(
            np.iinfo(np.int64).max + 1,
            np.dtype("uint64"),
            id="max_np.int64_plus_one_cols-expected_uint64_on_disk",
            marks=pytest.mark.xfail(
                reason="scipy sparse does not support bigger than max(int64) values in indices and there is no uint128."
            ),
        ),
        pytest.param(
            np.iinfo(np.uint64).max + 1,
            np.dtype("uint64"),
            id="max_np.uint64_plus_one_cols-expected_uint64_on_disk",
            marks=pytest.mark.xfail(
                reason="scipy sparse does not support bigger than max(int64) values in indices and there is no uint128."
            ),
        ),
    ],
)
@pytest.mark.parametrize("format", ["csr", "csc"])
def test_write_indices_min(
    store: H5Group | ZarrGroup,
    num_minor_axis: int,
    expected_dtype: np.dtype,
    format: Literal["csr", "csc"],
):
    minor_axis_index = np.array([num_minor_axis - 1])
    major_axis_index = np.array([10])
    row_cols = (
        (minor_axis_index, major_axis_index)
        if format == "csc"
        else (major_axis_index, minor_axis_index)
    )
    shape = (num_minor_axis, 20) if format == "csc" else (20, num_minor_axis)
    X = getattr(sparse, f"{format}_array")(
        (np.array([10]), row_cols),
        shape=shape,
    )
    assert X.nnz == 1
    with ad.settings.override(write_csr_csc_indices_with_min_possible_dtype=True):
        write_elem(store, "X", X)

    assert store["X/indices"].dtype == expected_dtype
    with ad.settings.override(use_sparse_array_on_read=True):
        result = read_elem(store["X"])
    assert_equal(result.data, X.data)
    assert_equal(result.indices, X.indices)
    assert_equal(result.indptr, X.indptr)
    assert X.format == result.format
    assert result.shape == X.shape
    # != comparison converts to csr, which allocates a lot of memory or errors out with:
    # ValueError: array is too big; `arr.size * arr.dtype.itemsize` is larger than the maximum possible size.
    # Because the old, very large, minor axis is now the major axis and so either it fails to create or the indptr is very big.
    # The above tests should be enough to capture the desired equality checks so this is mostly for being extra sure.
    # See https://github.com/scipy/scipy/issues/23826
    if not (format == "csc" and num_minor_axis > np.iinfo(np.uint16).max + 1):
        assert (result != X).nnz == 0


def test_io_spec_raw(store):
    adata = gen_adata((3, 2), **GEN_ADATA_NO_XARRAY_ARGS)
    adata.raw = adata.copy()

    write_elem(store, "adata", adata)

    assert _read_attr(store["adata/raw"].attrs, "encoding-type") == "raw"

    from_disk = read_elem(store["adata"])
    assert_equal(from_disk.raw, adata.raw)


def test_write_anndata_to_root(store):
    adata = gen_adata((3, 2), **GEN_ADATA_NO_XARRAY_ARGS)

    write_elem(store, "/", adata)
    # TODO: see https://github.com/zarr-developers/zarr-python/issues/2716
    if not is_zarr_v2() and isinstance(store, ZarrGroup):
        store = zarr.open(store.store)
    from_disk = read_elem(store)

    assert _read_attr(store.attrs, "encoding-type") == "anndata"
    assert_equal(from_disk, adata)


@pytest.mark.parametrize(
    ("attribute", "value"),
    [
        ("encoding-type", "floob"),
        ("encoding-version", "10000.0"),
    ],
)
def test_read_iospec_not_found(store, attribute, value):
    adata = gen_adata((3, 2), **GEN_ADATA_NO_XARRAY_ARGS)

    write_elem(store, "/", adata)
    store["obs"].attrs.update({attribute: value})

    with pytest.raises(IORegistryError) as exc_info:
        read_elem(store)
    msg = str(exc_info.value)

    assert "No read method registered for IOSpec" in msg
    assert f"{attribute.replace('-', '_')}='{value}'" in msg


@pytest.mark.parametrize(
    "obj",
    [(b"x",)],
)
def test_write_io_error(store, obj):
    full_pattern = re.compile(
        rf"No method registered for writing {type(obj)} into .*Group"
    )

    with pytest.raises(IORegistryError, match=r"while writing key '/el'") as exc_info:
        write_elem(store, "/el", obj)

    msg = str(exc_info.value)
    assert re.search(full_pattern, msg)


@pytest.mark.parametrize(
    ("ad_setting", "pd_setting", "expected_missing", "expected_no_missing"),
    [
        # when explicitly disabled, we expect an error when trying to write an array with missing values
        # and expect an array without missing values to be written in the old, non-nullable format
        *(
            pytest.param(
                False,
                pd_ignored,
                *(
                    [
                        (
                            RuntimeError,
                            r"`anndata.settings.allow_write_nullable_strings` is False",
                        )
                    ]
                    * 2
                ),
                id=f"off-explicit-{int(pd_ignored)}",
            )
            for pd_ignored in [False, True]
        ),
        # when enabled, we expect arrays to be written in the nullable format
        pytest.param(True, False, *(["nullable-string-array"] * 2), id="on-explicit-0"),
        pytest.param(True, True, *(["nullable-string-array"] * 2), id="on-explicit-1"),
    ],
)
@pytest.mark.parametrize("missing", [True, False], ids=["missing", "no_missing"])
def test_write_nullable_string(
    *,
    store: GroupStorageType,
    ad_setting: bool,
    pd_setting: bool,
    expected_missing: tuple[type[Exception], str] | str,
    expected_no_missing: tuple[type[Exception], str] | str,
    missing: bool,
) -> None:
    expected = expected_missing if missing else expected_no_missing
    with (
        ad.settings.override(allow_write_nullable_strings=ad_setting),
        pd.option_context("future.infer_string", pd_setting),
        (
            nullcontext()
            if isinstance(expected, str)
            else pytest.raises(expected[0], match=expected[1])
        ),
    ):
        write_elem(store, "/el", pd.array([pd.NA if missing else "a"], dtype="string"))

    if isinstance(expected, str):
        assert store["el"].attrs["encoding-type"] == expected


@pytest.mark.parametrize("infer_string", [True, False], ids=["infer_string", "default"])
@pytest.mark.parametrize(
    "index",
    [
        pd.array([1, 2, pd.NA]),
        pd.array([1, 2, np.nan]),
        pd.RangeIndex(3),
    ],
    ids=["NA", "nan", "range"],
)
def test_nullable_string_from_non_string_index(
    *, infer_string: bool, index: pd.api.extensions.ExtensionArray
) -> None:
    with pd.option_context("future.infer_string", infer_string):
        adata = ad.AnnData(obs=pd.DataFrame({"foo": [1, 2, 3]}, index=index))
        assert ("string" if infer_string else object) == adata.obs_names.dtype


def test_categorical_order_type(store):
    # https://github.com/scverse/anndata/issues/853
    cat = pd.Categorical([0, 1], ordered=True)
    write_elem(store, "ordered", cat)
    write_elem(store, "unordered", cat.set_ordered(False))

    assert isinstance(read_elem(store["ordered"]).ordered, bool)
    assert read_elem(store["ordered"]).ordered is True
    assert isinstance(read_elem(store["unordered"]).ordered, bool)
    assert read_elem(store["unordered"]).ordered is False


def test_override_specification():
    """
    Test that trying to overwrite an existing encoding raises an error.
    """
    from copy import deepcopy

    registry = deepcopy(_REGISTRY)

    with pytest.raises(TypeError):

        @registry.register_write(
            ZarrGroup, ad.AnnData, IOSpec("some new type", "0.1.0")
        )
        def _(store, key, adata):
            pass


@pytest.mark.parametrize(
    "value",
    [
        pytest.param({"a": 1}, id="dict"),
        pytest.param(
            lambda: gen_adata((3, 2), **GEN_ADATA_NO_XARRAY_ARGS), id="anndata"
        ),
        pytest.param(sparse.random(5, 3, format="csr", density=0.5), id="csr_matrix"),
        pytest.param(sparse.random(5, 3, format="csc", density=0.5), id="csc_matrix"),
        pytest.param(pd.DataFrame({"a": [1, 2, 3]}), id="dataframe"),
        pytest.param(pd.Categorical(list("aabccedd")), id="categorical"),
        pytest.param(
            pd.Categorical(list("aabccedd"), ordered=True), id="categorical-ordered"
        ),
        pytest.param(
            pd.Categorical([1, 2, 1, 3], ordered=True), id="categorical-numeric"
        ),
        pytest.param(
            pd.arrays.IntegerArray(
                np.ones(5, dtype=int), mask=np.array([True, False, True, False, True])
            ),
            id="nullable-integer",
        ),
        pytest.param(pd.array([1, 2, 3]), id="nullable-integer-no-nulls"),
        pytest.param(
            pd.arrays.BooleanArray(
                np.random.randint(0, 2, size=5, dtype=bool),
                mask=np.random.randint(0, 2, size=5, dtype=bool),
            ),
            id="nullable-boolean",
        ),
        pytest.param(
            pd.array([True, False, True, True]), id="nullable-boolean-no-nulls"
        ),
    ],
)
def test_write_to_root(store: GroupStorageType, value):
    """
    Test that elements which are written as groups can we written to the root group.
    """
    if callable(value):
        value = value()
    write_elem(store, "/", value)
    # See: https://github.com/zarr-developers/zarr-python/issues/2716
    if isinstance(store, ZarrGroup) and not is_zarr_v2():
        store = zarr.open(store.store)
    result = read_elem(store)

    assert_equal(result, value)


@pytest.mark.parametrize("consolidated", [True, False])
@pytest.mark.zarr_io
def test_read_zarr_from_group(tmp_path, consolidated):
    # https://github.com/scverse/anndata/issues/1056
    pth = tmp_path / "test.zarr"
    adata = gen_adata((3, 2), **GEN_ADATA_NO_XARRAY_ARGS)

    z = open_write_group(pth)
    write_elem(z, "table/table", adata)

    if consolidated:
        zarr.consolidate_metadata(z.store)

    read_func = zarr.open_consolidated if consolidated else zarr.open

    z = read_func(pth)
    expected = ad.read_zarr(z["table/table"])
    assert_equal(adata, expected)


def test_dataframe_column_uniqueness(store):
    repeated_cols = pd.DataFrame(np.ones((3, 2)), columns=["a", "a"])

    with pytest.raises(
        ValueError,
        match=r"Found repeated column names: \['a'\]\. Column names must be unique\.",
    ):
        write_elem(store, "repeated_cols", repeated_cols)

    index_shares_col_name = pd.DataFrame(
        {"col_name": [1, 2, 3]}, index=pd.Index([1, 3, 2], name="col_name")
    )

    with pytest.raises(
        ValueError,
        match=r"DataFrame\.index\.name \('col_name'\) is also used by a column whose values are different\.",
    ):
        write_elem(store, "index_shares_col_name", index_shares_col_name)

    index_shared_okay = pd.DataFrame(
        {"col_name": [1, 2, 3]}, index=pd.Index([1, 2, 3], name="col_name")
    )

    write_elem(store, "index_shared_okay", index_shared_okay)
    result = read_elem(store["index_shared_okay"])

    assert_equal(result, index_shared_okay)


@pytest.mark.parametrize(
    "copy_on_write",
    [
        pytest.param(True, id="cow"),
        pytest.param(
            False,
            marks=pytest.mark.skipif(
                PANDAS_3, reason="Can’t disable copy-on-write in pandas 3+."
            ),
            id="nocow",
        ),
    ],
)
def test_io_pd_cow(
    *, exit_stack: ExitStack, store: GroupStorageType, copy_on_write: bool
) -> None:
    """See <https://github.com/zarr-developers/numcodecs/issues/514>."""
    if not PANDAS_3:  # Setting copy_on_write always warns in pandas 3, and does nothing
        exit_stack.enter_context(pd.option_context("mode.copy_on_write", copy_on_write))
    orig = gen_adata((3, 2), **GEN_ADATA_NO_XARRAY_ARGS)
    write_elem(store, "adata", orig)
    from_store = read_elem(store["adata"])
    assert_equal(orig, from_store)


def test_read_sparse_array(
    tmp_path: Path,
    sparse_format: Literal["csr", "csc"],
    diskfmt: Literal["h5ad", "zarr"],
):
    path = tmp_path / f"test.{diskfmt.replace('ad', '')}"
    a = sparse.random(100, 100, format=sparse_format)
    f = open_write_group(path, mode="a") if diskfmt == "zarr" else h5py.File(path, "a")
    ad.io.write_elem(f, "mtx", a)
    ad.settings.use_sparse_array_on_read = True
    mtx = ad.io.read_elem(f["mtx"])
    assert issubclass(type(mtx), CSArray)


@pytest.mark.parametrize(
    ("chunks", "expected_chunks"),
    [((1,), (1,)), ((-1,), (120,)), (None, (25,))],
    ids=["small", "minus_one_uses_full", "none_uses_ondisk_chunking"],
)
@pytest.mark.parametrize(
    "arr", [np.arange(120), np.array(["a"] * 120)], ids=["numeric", "string"]
)
def test_chunking_1d_array(
    store: H5Group | ZarrGroup,
    arr: np.ndarray,
    chunks: tuple[int] | None,
    expected_chunks: tuple[int],
):
    write_elem(store, "foo", arr, dataset_kwargs={"chunks": 25})
    arr = read_elem_lazy(store["foo"], chunks=chunks)
    assert arr.chunksize == expected_chunks


@pytest.mark.parametrize(
    ("chunks", "expected_chunks"),
    [
        ((1, 50), (1, 50)),
        ((10, -1), (10, 50)),
        ((10, None), (10, 50)),
        (None, (25, 25)),
    ],
    ids=[
        "small",
        "minus_one_uses_full",
        "none_on_axis_uses_full",
        "none_uses_ondisk_chunking",
    ],
)
def test_chunking_2d_array(
    store: H5Group | ZarrGroup,
    chunks: tuple[int] | None,
    expected_chunks: tuple[int],
):
    write_elem(
        store,
        "foo",
        np.arange(100 * 50).reshape((100, 50)),
        dataset_kwargs={"chunks": (25, 25)},
    )
    arr = read_elem_lazy(store["foo"], chunks=chunks)
    assert arr.chunksize == expected_chunks


@pytest.mark.parametrize(
    ("shape", "expected_chunks"),
    [((100, 50), (100, 50)), ((1100, 1100), (1000, 1000))],
    ids=["under_default", "over_default"],
)
def test_h5_unchunked(
    shape: tuple[int, int],
    expected_chunks: tuple[int, int],
    tmp_path: Path,
):
    with h5py.File(tmp_path / "foo.h5ad", mode="w") as f:
        write_elem(
            f,
            "foo",
            np.arange(shape[0] * shape[1]).reshape(shape),
        )
        arr = read_elem_lazy(f["foo"])
    assert arr.chunksize == expected_chunks


@pytest.mark.zarr_io
def test_write_auto_sharded(tmp_path: Path):
    if is_zarr_v2():
        with (
            pytest.raises(
                ValueError, match=r"Cannot use sharding with `zarr-python<3`."
            ),
            ad.settings.override(auto_shard_zarr_v3=True),
        ):
            pass
    else:
        path = tmp_path / "check.zarr"
        adata = gen_adata((1000, 100), **GEN_ADATA_NO_XARRAY_ARGS)
        with ad.settings.override(auto_shard_zarr_v3=True, zarr_write_format=3):
            adata.write_zarr(path)

        check_all_sharded(zarr.open(path))


@pytest.mark.zarr_io
@pytest.mark.skipif(is_zarr_v2(), reason="auto sharding is allowed only for zarr v3.")
def test_write_auto_sharded_against_v2_format():
    with pytest.raises(ValueError, match=r"Cannot shard v2 format data."):  # noqa: PT012, SIM117
        with ad.settings.override(zarr_write_format=2):
            with ad.settings.override(auto_shard_zarr_v3=True):
                pass


@pytest.mark.zarr_io
@pytest.mark.skipif(is_zarr_v2(), reason="auto sharding is allowed only for zarr v3.")
def test_write_auto_cannot_set_v2_format_after_sharding():
    with pytest.raises(ValueError, match=r"Cannot set `zarr_write_format` to 2"):  # noqa: PT012, SIM117
        with ad.settings.override(zarr_write_format=3):
            with ad.settings.override(auto_shard_zarr_v3=True):
                with ad.settings.override(zarr_write_format=2):
                    pass


@pytest.mark.zarr_io
@pytest.mark.skipif(is_zarr_v2(), reason="auto sharding is allowed only for zarr v3.")
def test_write_auto_sharded_does_not_override(tmp_path: Path):
    z = open_write_group(tmp_path / "arr.zarr", zarr_format=3)
    X = sparse.random(
        100, 100, density=0.1, format="csr", rng=np.random.default_rng(42)
    )
    with ad.settings.override(auto_shard_zarr_v3=True, zarr_write_format=3):
        ad.io.write_elem(z, "X_default", X)
        shards_default = z["X_default"]["indices"].shards
        new_shards = shards_default[0] // 2
        new_shards = int(new_shards - new_shards % 2)
        ad.io.write_elem(
            z,
            "X_manually_set",
            X,
            dataset_kwargs={
                "shards": (new_shards,),
                "chunks": (int(new_shards / 2),),
            },
        )

    def visitor(key: str, array: zarr.Array):
        assert array.shards == (new_shards,)

    visititems_zarr(z["X_manually_set"], visitor)



from __future__ import annotations

import warnings
from collections.abc import Hashable
from copy import deepcopy
from functools import partial, singledispatch
from importlib.metadata import version
from importlib.util import find_spec
from itertools import chain, permutations, product
from operator import attrgetter
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import pytest
from boltons.iterutils import default_exit, remap, research
from numpy import ma
from packaging.version import Version
from scipy import sparse

from anndata import AnnData, Raw, concat
from anndata._core import merge
from anndata._core.index import _subset
from anndata._core.xarray import Dataset2D
from anndata.compat import AwkArray, CSArray, CSMatrix, CupySparseMatrix, DaskArray
from anndata.tests import helpers
from anndata.tests.helpers import (
    BASE_MATRIX_PARAMS,
    CUPY_MATRIX_PARAMS,
    DASK_MATRIX_PARAMS,
    DEFAULT_COL_TYPES,
    GEN_ADATA_DASK_ARGS,
    as_dense_dask_array,
    assert_equal,
    gen_adata,
    gen_vstr_recarray,
)
from anndata.utils import asarray

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Literal

mark_legacy_concatenate = pytest.mark.filterwarnings(
    r"ignore:.*AnnData\.concatenate is deprecated:FutureWarning"
)


@singledispatch
def filled_like(a, fill_value=None):
    raise NotImplementedError()


@filled_like.register(np.ndarray)
def _filled_array_np(a, fill_value=None):
    if fill_value is None:
        fill_value = np.nan
    return np.broadcast_to(fill_value, a.shape)


@filled_like.register(DaskArray)
def _filled_array(a, fill_value=None):
    return as_dense_dask_array(_filled_array_np(a, fill_value))


@filled_like.register(CSMatrix)
def _filled_sparse(a, fill_value=None):
    if fill_value is None:
        return sparse.csr_matrix(a.shape)
    else:
        return sparse.csr_matrix(np.broadcast_to(fill_value, a.shape))


@filled_like.register(CSArray)
def _filled_sparse_array(a, fill_value=None):
    return sparse.csr_array(filled_like(sparse.csr_matrix(a)))


@filled_like.register(pd.DataFrame)
def _filled_df(a, fill_value=np.nan):
    # dtype from pd.concat can be unintuitive, this returns something close enough
    return a.loc[[], :].reindex(index=a.index, fill_value=fill_value)


def check_filled_like(x, fill_value=None, elem_name=None):
    if fill_value is None:
        assert_equal(x, filled_like(x), elem_name=elem_name)
    else:
        assert_equal(x, filled_like(x, fill_value=fill_value), elem_name=elem_name)


def make_idx_tuple(idx, axis):
    tup = [slice(None), slice(None)]
    tup[axis] = idx
    return tuple(tup)


# Will call func(sparse_matrix) so these types should be sparse compatible
# See array_type if only dense arrays are expected as input.
@pytest.fixture(params=BASE_MATRIX_PARAMS + DASK_MATRIX_PARAMS + CUPY_MATRIX_PARAMS)
def array_type(request):
    return request.param


@pytest.fixture(params=BASE_MATRIX_PARAMS + DASK_MATRIX_PARAMS)
def cpu_array_type(request):
    return request.param


@pytest.fixture(params=["inner", "outer"])
def join_type(request):
    return request.param


@pytest.fixture(params=[0, np.nan, np.pi])
def fill_val(request):
    return request.param


@pytest.fixture(params=["obs", "var"])
def axis_name(request) -> Literal["obs", "var"]:
    return request.param


@pytest.fixture(params=list(merge.MERGE_STRATEGIES.keys()))
def merge_strategy(request):
    return request.param


@pytest.fixture(
    params=[
        pytest.param(False, id="pandas"),
        pytest.param(
            True,
            marks=pytest.mark.skipif(
                not find_spec("xarray"), reason="xarray not installed."
            ),
            id="xarray",
        ),
    ],
)
def use_xdataset(request):
    return request.param


@pytest.fixture(
    params=[
        pytest.param(False, id="concat-in-memory"),
        pytest.param(True, id="concat-lazy"),
    ],
)
def force_lazy(request):
    return request.param


def fix_known_differences(
    orig: AnnData, result: AnnData, *, backwards_compat: bool = True
):
    """
    Helper function for reducing anndata's to only the elements we expect to be
    equivalent after concatenation.

    Only for the case where orig is the ground truth result of what concatenation should be.

    If backwards_compat, checks against what `AnnData.concatenate` could do. Otherwise checks for `concat`.
    """
    orig = orig.copy()
    result = result.copy()

    if backwards_compat:
        del orig.varm
        del orig.varp
        if isinstance(result.obs, Dataset2D):
            result.obs = result.obs.ds.drop_vars(["batch"])
        else:
            result.obs.drop(columns=["batch"], inplace=True)

    for attrname in ("obs", "var"):
        if isinstance(getattr(result, attrname), Dataset2D):
            for adata in (orig, result):
                df = getattr(adata, attrname).ds.to_dataframe()[
                    getattr(orig, attrname).columns
                ]
                df.index.name = "index"
                setattr(adata, attrname, df)
            resattr = getattr(result, attrname)
            origattr = getattr(orig, attrname)
            for colname, col in resattr.items():
                # concatenation of XDatasets happens via Dask arrays and those don't know about Pandas Extension arrays
                # so categoricals and nullable arrays are all converted to other dtypes
                if col.dtype != origattr[
                    colname
                ].dtype and pd.api.types.is_extension_array_dtype(
                    origattr[colname].dtype
                ):
                    resattr[colname] = col.astype(origattr[colname].dtype)

    result.strings_to_categoricals()  # Should this be implicit in concatenation?

    # TODO
    # * merge varm, varp similar to uns
    # * merge obsp, but some information should be lost
    del orig.obsp  # TODO

    # Possibly need to fix this, ordered categoricals lose orderedness
    for get_df in [lambda k: k.obs, lambda k: k.obsm["df"]]:
        str_to_df_converted = get_df(result)
        for k, dtype in get_df(orig).dtypes.items():
            if isinstance(dtype, pd.CategoricalDtype) and dtype.ordered:
                str_to_df_converted[k] = str_to_df_converted[k].astype(dtype)

    return orig, result


def test_concat_interface_errors(use_xdataset):
    adatas = [
        gen_adata((5, 10), obs_xdataset=use_xdataset, var_xdataset=use_xdataset),
        gen_adata((5, 10), obs_xdataset=use_xdataset, var_xdataset=use_xdataset),
    ]

    with pytest.raises(ValueError, match=r"`axis` must be.*0, 1, 'obs', or 'var'"):
        concat(adatas, axis=3)
    with pytest.raises(ValueError, match="'inner' or 'outer'"):
        concat(adatas, join="not implemented")
    with pytest.raises(ValueError, match="No objects to concatenate"):
        concat([])


@mark_legacy_concatenate
@pytest.mark.parametrize(
    ("concat_func", "backwards_compat"),
    [
        (partial(concat, merge="unique"), False),
        (lambda x, **kwargs: x[0].concatenate(x[1:], **kwargs), True),
    ],
)
def test_concatenate_roundtrip(
    join_type,
    array_type,
    concat_func,
    backwards_compat,
    use_xdataset,
    force_lazy,
):
    if backwards_compat and force_lazy:
        pytest.skip("unsupported")
    adata = gen_adata(
        (100, 10),
        X_type=array_type,
        obs_xdataset=use_xdataset,
        var_xdataset=use_xdataset,
        **GEN_ADATA_DASK_ARGS,
    )

    remaining = adata.obs_names
    subsets = []
    while len(remaining) > 0:
        n = min(len(remaining), np.random.choice(50))
        subset_idx = np.random.choice(remaining, n, replace=False)
        subsets.append(adata[subset_idx])
        remaining = remaining.difference(subset_idx)
    result = concat_func(subsets, join=join_type, uns_merge="same", index_unique=None)
    if backwards_compat and use_xdataset:
        import xarray as xr

        result.var = xr.Dataset.from_dataframe(
            result.var
        )  # backwards compat always returns a dataframe

    # Correcting for known differences
    orig, result = fix_known_differences(
        adata, result, backwards_compat=backwards_compat
    )

    assert_equal(result[orig.obs_names].copy(), orig)
    base_type = type(orig.X)
    if sparse.issparse(orig.X):
        base_type = CSArray if isinstance(orig.X, CSArray) else CSMatrix
    if isinstance(orig.X, CupySparseMatrix):
        base_type = CupySparseMatrix
    assert isinstance(result.X, base_type)


@mark_legacy_concatenate
def test_concatenate_dense():
    # dense data
    X1 = np.array([[1, 2, 3], [4, 5, 6]])
    X2 = np.array([[1, 2, 3], [4, 5, 6]])
    X3 = np.array([[1, 2, 3], [4, 5, 6]])

    adata1 = AnnData(
        X1,
        dict(obs_names=["s1", "s2"], anno1=["c1", "c2"]),
        dict(var_names=["a", "b", "c"], annoA=[0, 1, 2]),
        obsm=dict(X_1=X1, X_2=X2, X_3=X3),
        layers=dict(Xs=X1),
    )
    adata2 = AnnData(
        X2,
        dict(obs_names=["s3", "s4"], anno1=["c3", "c4"]),
        dict(var_names=["d", "c", "b"], annoA=[0, 1, 2]),
        obsm=dict(X_1=X1, X_2=X2, X_3=X3),
        layers={"Xs": X2},
    )
    adata3 = AnnData(
        X3,
        dict(obs_names=["s1", "s2"], anno2=["d3", "d4"]),
        dict(var_names=["d", "c", "b"], annoB=[0, 1, 2]),
        obsm=dict(X_1=X1, X_2=X2),
        layers=dict(Xs=X3),
    )

    # inner join
    adata = adata1.concatenate(adata2, adata3)
    X_combined = [[2, 3], [5, 6], [3, 2], [6, 5], [3, 2], [6, 5]]
    assert adata.X.astype(int).tolist() == X_combined
    assert adata.layers["Xs"].astype(int).tolist() == X_combined
    assert adata.obs.columns.tolist() == ["anno1", "anno2", "batch"]
    assert adata.var.columns.tolist() == ["annoA-0", "annoA-1", "annoB-2"]
    assert adata.var.values.tolist() == [[1, 2, 2], [2, 1, 1]]
    assert adata.obsm.keys() == {"X_1", "X_2"}
    assert adata.obsm["X_1"].tolist() == np.concatenate([X1, X1, X1]).tolist()

    # with batch_key and batch_categories
    adata = adata1.concatenate(adata2, adata3, batch_key="batch1")
    assert adata.obs.columns.tolist() == ["anno1", "anno2", "batch1"]
    adata = adata1.concatenate(adata2, adata3, batch_categories=["a1", "a2", "a3"])
    assert adata.obs["batch"].cat.categories.tolist() == ["a1", "a2", "a3"]
    assert adata.var_names.tolist() == ["b", "c"]

    # outer join
    adata = adata1.concatenate(adata2, adata3, join="outer")

    X_ref = np.array([
        [1.0, 2.0, 3.0, np.nan],
        [4.0, 5.0, 6.0, np.nan],
        [np.nan, 3.0, 2.0, 1.0],
        [np.nan, 6.0, 5.0, 4.0],
        [np.nan, 3.0, 2.0, 1.0],
        [np.nan, 6.0, 5.0, 4.0],
    ])
    np.testing.assert_equal(adata.X, X_ref)
    var_ma = ma.masked_invalid(adata.var.values.tolist())
    var_ma_ref = ma.masked_invalid(
        np.array([
            [0.0, np.nan, np.nan],
            [1.0, 2.0, 2.0],
            [2.0, 1.0, 1.0],
            [np.nan, 0.0, 0.0],
        ])
    )
    assert np.array_equal(var_ma.mask, var_ma_ref.mask)
    assert np.allclose(var_ma.compressed(), var_ma_ref.compressed())


@mark_legacy_concatenate
def test_concatenate_layers(array_type, join_type):
    adatas = []
    for _ in range(5):
        a = array_type(sparse.random(100, 200, format="csr"))
        adatas.append(AnnData(X=a, layers={"a": a}))

    merged = adatas[0].concatenate(adatas[1:], join=join_type)
    assert_equal(merged.X, merged.layers["a"])


@pytest.fixture
def obsm_adatas():
    def gen_index(n):
        return [f"cell{i}" for i in range(n)]

    return [
        AnnData(
            X=sparse.csr_matrix((3, 5)),
            obs=pd.DataFrame(index=gen_index(3)),
            obsm={
                "dense": np.arange(6).reshape(3, 2),
                "sparse": sparse.csr_matrix(np.arange(6).reshape(3, 2)),
                "df": pd.DataFrame(
                    {
                        "a": np.arange(3),
                        "b": list("abc"),
                        "c": pd.Categorical(list("aab")),
                    },
                    index=gen_index(3),
                ),
            },
        ),
        AnnData(
            X=sparse.csr_matrix((4, 10)),
            obs=pd.DataFrame(index=gen_index(4)),
            obsm=dict(
                dense=np.arange(12).reshape(4, 3),
                df=pd.DataFrame(dict(a=np.arange(3, 7)), index=gen_index(4)),
            ),
        ),
        AnnData(
            X=sparse.csr_matrix((2, 100)),
            obs=pd.DataFrame(index=gen_index(2)),
            obsm={
                "sparse": np.arange(8).reshape(2, 4),
                "dense": np.arange(4, 8).reshape(2, 2),
                "df": pd.DataFrame(
                    {
                        "a": np.arange(7, 9),
                        "b": list("cd"),
                        "c": pd.Categorical(list("ab")),
                    },
                    index=gen_index(2),
                ),
            },
        ),
    ]


@mark_legacy_concatenate
def test_concatenate_obsm_inner(obsm_adatas):
    adata = obsm_adatas[0].concatenate(obsm_adatas[1:], join="inner")

    assert set(adata.obsm.keys()) == {"dense", "df"}
    assert adata.obsm["dense"].shape == (9, 2)
    assert adata.obsm["dense"].tolist() == [
        [0, 1],
        [2, 3],
        [4, 5],
        [0, 1],
        [3, 4],
        [6, 7],
        [9, 10],
        [4, 5],
        [6, 7],
    ]

    assert adata.obsm["df"].columns == ["a"]
    assert adata.obsm["df"]["a"].tolist() == list(range(9))
    # fmt: off
    true_df = (
        pd.concat([a.obsm["df"] for a in obsm_adatas], join="inner")
        .reset_index(drop=True)
    )
    # fmt: on
    cur_df = adata.obsm["df"].reset_index(drop=True)
    pd.testing.assert_frame_equal(true_df, cur_df)


@mark_legacy_concatenate
def test_concatenate_obsm_outer(obsm_adatas, fill_val):
    outer = obsm_adatas[0].concatenate(
        obsm_adatas[1:], join="outer", fill_value=fill_val
    )

    inner = obsm_adatas[0].concatenate(obsm_adatas[1:], join="inner")
    for k, inner_v in inner.obsm.items():
        assert np.array_equal(
            _subset(outer.obsm[k], (slice(None), slice(None, inner_v.shape[1]))),
            inner_v,
        )

    assert set(outer.obsm.keys()) == {"dense", "df", "sparse"}
    assert isinstance(outer.obsm["dense"], np.ndarray)
    np.testing.assert_equal(
        outer.obsm["dense"],
        np.array([
            [0, 1, fill_val],
            [2, 3, fill_val],
            [4, 5, fill_val],
            [0, 1, 2],
            [3, 4, 5],
            [6, 7, 8],
            [9, 10, 11],
            [4, 5, fill_val],
            [6, 7, fill_val],
        ]),
    )

    assert isinstance(outer.obsm["sparse"], CSMatrix)
    np.testing.assert_equal(
        outer.obsm["sparse"].toarray(),
        np.array([
            [0, 1, fill_val, fill_val],
            [2, 3, fill_val, fill_val],
            [4, 5, fill_val, fill_val],
            [fill_val, fill_val, fill_val, fill_val],
            [fill_val, fill_val, fill_val, fill_val],
            [fill_val, fill_val, fill_val, fill_val],
            [fill_val, fill_val, fill_val, fill_val],
            [0, 1, 2, 3],
            [4, 5, 6, 7],
        ]),
    )

    # fmt: off
    true_df = (
        pd.concat([a.obsm["df"] for a in obsm_adatas], join="outer")
        .reset_index(drop=True)
    )
    # fmt: on
    cur_df = outer.obsm["df"].reset_index(drop=True)
    pd.testing.assert_frame_equal(true_df, cur_df)


@pytest.mark.parametrize(
    ("axis", "axis_name"),
    [("obs", 0), ("var", 1)],
)
def test_concat_axis_param(axis, axis_name):
    a, b = gen_adata((10, 10)), gen_adata((10, 10))
    assert_equal(concat([a, b], axis=axis), concat([a, b], axis=axis_name))


def test_concat_annot_join(obsm_adatas, join_type):
    adatas = [
        AnnData(sparse.csr_matrix(a.shape), obs=a.obsm["df"], var=a.var)
        for a in obsm_adatas
    ]
    pd.testing.assert_frame_equal(
        concat(adatas, join=join_type).obs,
        pd.concat([a.obs for a in adatas], join=join_type),
    )


@mark_legacy_concatenate
def test_concatenate_layers_misaligned(array_type, join_type):
    adatas = []
    for _ in range(5):
        a = array_type(sparse.random(100, 200, format="csr"))
        adata = AnnData(X=a, layers={"a": a})
        adatas.append(
            adata[:, np.random.choice(adata.var_names, 150, replace=False)].copy()
        )

    merged = adatas[0].concatenate(adatas[1:], join=join_type)
    assert_equal(merged.X, merged.layers["a"])


@mark_legacy_concatenate
def test_concatenate_layers_outer(array_type, fill_val):
    # Testing that issue #368 is fixed
    a = AnnData(
        X=np.ones((10, 20)),
        layers={"a": array_type(sparse.random(10, 20, format="csr"))},
    )
    b = AnnData(X=np.ones((10, 20)))

    c = a.concatenate(b, join="outer", fill_value=fill_val, batch_categories=["a", "b"])

    np.testing.assert_array_equal(
        asarray(c[c.obs["batch"] == "b"].layers["a"]), fill_val
    )


@mark_legacy_concatenate
def test_concatenate_fill_value(fill_val):
    def get_obs_els(adata):
        return {
            "X": adata.X,
            **{f"layer_{k}": adata.layers[k] for k in adata.layers},
            **{f"obsm_{k}": adata.obsm[k] for k in adata.obsm},
        }

    adata1 = gen_adata((10, 10))
    adata1.obsm = {
        k: v
        for k, v in adata1.obsm.items()
        if not isinstance(v, pd.DataFrame | AwkArray | Dataset2D)
    }
    adata2 = gen_adata((10, 5))
    adata2.obsm = {
        k: v[:, : v.shape[1] // 2]
        for k, v in adata2.obsm.items()
        if not isinstance(v, pd.DataFrame | AwkArray | Dataset2D)
    }
    adata3 = gen_adata((7, 3))
    adata3.obsm = {
        k: v[:, : v.shape[1] // 3]
        for k, v in adata3.obsm.items()
        if not isinstance(v, pd.DataFrame | AwkArray | Dataset2D)
    }
    # remove AwkArrays from adata.var, as outer joins are not yet implemented for them
    for tmp_ad in [adata1, adata2, adata3]:
        for k in [k for k, v in tmp_ad.varm.items() if isinstance(v, AwkArray)]:
            del tmp_ad.varm[k]

    joined = adata1.concatenate([adata2, adata3], join="outer", fill_value=fill_val)

    ptr = 0
    for orig in [adata1, adata2, adata3]:
        cur = joined[ptr : ptr + orig.n_obs]
        cur_els = get_obs_els(cur)
        orig_els = get_obs_els(orig)
        for k, cur_v in cur_els.items():
            orig_v = orig_els.get(k, sparse.csr_matrix((orig.n_obs, 0)))
            assert_equal(cur_v[:, : orig_v.shape[1]], orig_v)
            np.testing.assert_equal(asarray(cur_v[:, orig_v.shape[1] :]), fill_val)
        ptr += orig.n_obs


@mark_legacy_concatenate
def test_concatenate_dense_duplicates():
    X1 = np.array([[1, 2, 3], [4, 5, 6]])
    X2 = np.array([[1, 2, 3], [4, 5, 6]])
    X3 = np.array([[1, 2, 3], [4, 5, 6]])

    # inner join duplicates
    adata1 = AnnData(
        X1,
        dict(obs_names=["s1", "s2"], anno1=["c1", "c2"]),
        dict(
            var_names=["a", "b", "c"],
            annoA=[0, 1, 2],
            annoB=[1.1, 1.0, 2.0],
            annoC=[1.1, 1.0, 2.0],
            annoD=[2.1, 2.0, 3.0],
        ),
    )
    adata2 = AnnData(
        X2,
        dict(obs_names=["s3", "s4"], anno1=["c3", "c4"]),
        dict(
            var_names=["a", "b", "c"],
            annoA=[0, 1, 2],
            annoB=[1.1, 1.0, 2.0],
            annoC=[1.1, 1.0, 2.0],
            annoD=[2.1, 2.0, 3.0],
        ),
    )
    adata3 = AnnData(
        X3,
        dict(obs_names=["s1", "s2"], anno2=["d3", "d4"]),
        dict(
            var_names=["a", "b", "c"],
            annoA=[0, 1, 2],
            annoB=[1.1, 1.0, 2.0],
            annoD=[2.1, 2.0, 3.1],
        ),
    )

    adata = adata1.concatenate(adata2, adata3)
    assert adata.var.columns.tolist() == [
        "annoA",
        "annoB",
        "annoC-0",
        "annoD-0",
        "annoC-1",
        "annoD-1",
        "annoD-2",
    ]


@mark_legacy_concatenate
def test_concatenate_sparse():
    # sparse data
    from scipy.sparse import csr_matrix

    X1 = csr_matrix([[0, 2, 3], [0, 5, 6]])
    X2 = csr_matrix([[0, 2, 3], [0, 5, 6]])
    X3 = csr_matrix([[1, 2, 0], [0, 5, 6]])

    adata1 = AnnData(
        X1,
        dict(obs_names=["s1", "s2"], anno1=["c1", "c2"]),
        dict(var_names=["a", "b", "c"]),
        layers=dict(Xs=X1),
    )
    adata2 = AnnData(
        X2,
        dict(obs_names=["s3", "s4"], anno1=["c3", "c4"]),
        dict(var_names=["d", "c", "b"]),
        layers=dict(Xs=X2),
    )
    adata3 = AnnData(
        X3,
        dict(obs_names=["s5", "s6"], anno2=["d3", "d4"]),
        dict(var_names=["d", "c", "b"]),
        layers=dict(Xs=X3),
    )

    # inner join
    adata = adata1.concatenate(adata2, adata3)
    X_combined = [[2, 3], [5, 6], [3, 2], [6, 5], [0, 2], [6, 5]]
    assert adata.X.toarray().astype(int).tolist() == X_combined
    assert adata.layers["Xs"].toarray().astype(int).tolist() == X_combined

    # outer join
    adata = adata1.concatenate(adata2, adata3, join="outer")
    assert adata.X.toarray().tolist() == [
        [0.0, 2.0, 3.0, 0.0],
        [0.0, 5.0, 6.0, 0.0],
        [0.0, 3.0, 2.0, 0.0],
        [0.0, 6.0, 5.0, 0.0],
        [0.0, 0.0, 2.0, 1.0],
        [0.0, 6.0, 5.0, 0.0],
    ]


@mark_legacy_concatenate
def test_concatenate_mixed():
    X1 = sparse.csr_matrix(np.array([[1, 2, 0], [4, 0, 6], [0, 0, 9]]))
    X2 = sparse.csr_matrix(np.array([[0, 2, 3], [4, 0, 0], [7, 0, 9]]))
    X3 = sparse.csr_matrix(np.array([[1, 0, 3], [0, 0, 6], [0, 8, 0]]))
    X4 = np.array([[0, 2, 3], [4, 0, 0], [7, 0, 9]])
    adata1 = AnnData(
        X1,
        dict(obs_names=["s1", "s2", "s3"], anno1=["c1", "c2", "c3"]),
        dict(var_names=["a", "b", "c"], annoA=[0, 1, 2]),
        layers=dict(counts=X1),
    )
    adata2 = AnnData(
        X2,
        dict(obs_names=["s4", "s5", "s6"], anno1=["c3", "c4", "c5"]),
        dict(var_names=["d", "c", "b"], annoA=[0, 1, 2]),
        layers=dict(counts=X4),  # sic
    )
    adata3 = AnnData(
        X3,
        dict(obs_names=["s7", "s8", "s9"], anno2=["d3", "d4", "d5"]),
        dict(var_names=["d", "c", "b"], annoA=[0, 2, 3], annoB=[0, 1, 2]),
        layers=dict(counts=X3),
    )
    adata4 = AnnData(
        X4,
        dict(obs_names=["s4", "s5", "s6"], anno1=["c3", "c4", "c5"]),
        dict(var_names=["d", "c", "b"], annoA=[0, 1, 2]),
        layers=dict(counts=X2),  # sic
    )

    adata_all = AnnData.concatenate(adata1, adata2, adata3, adata4)
    assert isinstance(adata_all.X, sparse.csr_matrix)
    assert isinstance(adata_all.layers["counts"], sparse.csr_matrix)


@mark_legacy_concatenate
def test_concatenate_with_raw():
    # dense data
    X1 = np.array([[1, 2, 3], [4, 5, 6]])
    X2 = np.array([[1, 2, 3], [4, 5, 6]])
    X3 = np.array([[1, 2, 3], [4, 5, 6]])

    X4 = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])

    adata1 = AnnData(
        X1,
        dict(obs_names=["s1", "s2"], anno1=["c1", "c2"]),
        dict(var_names=["a", "b", "c"], annoA=[0, 1, 2]),
        layers=dict(Xs=X1),
    )
    adata2 = AnnData(
        X2,
        dict(obs_names=["s3", "s4"], anno1=["c3", "c4"]),
        dict(var_names=["d", "c", "b"], annoA=[0, 1, 2]),
        layers=dict(Xs=X2),
    )
    adata3 = AnnData(
        X3,
        dict(obs_names=["s1", "s2"], anno2=["d3", "d4"]),
        dict(var_names=["d", "c", "b"], annoB=[0, 1, 2]),
        layers=dict(Xs=X3),
    )

    adata4 = AnnData(
        X4,
        dict(obs_names=["s1", "s2"], anno1=["c1", "c2"]),
        dict(var_names=["a", "b", "c", "z"], annoA=[0, 1, 2, 3]),
        layers=dict(Xs=X4),
    )

    adata1.raw = adata1.copy()
    adata2.raw = adata2.copy()
    adata3.raw = adata3.copy()

    adata_all = AnnData.concatenate(adata1, adata2, adata3)
    assert isinstance(adata_all.raw, Raw)
    assert set(adata_all.raw.var_names) == {"b", "c"}
    assert_equal(adata_all.raw.to_adata().obs, adata_all.obs)
    assert np.array_equal(adata_all.raw.X, adata_all.X)

    adata_all = AnnData.concatenate(adata1, adata2, adata3, join="outer")
    assert isinstance(adata_all.raw, Raw)
    assert set(adata_all.raw.var_names) == set("abcd")
    assert_equal(adata_all.raw.to_adata().obs, adata_all.obs)
    assert np.array_equal(np.nan_to_num(adata_all.raw.X), np.nan_to_num(adata_all.X))

    adata3.raw = adata4.copy()
    adata_all = AnnData.concatenate(adata1, adata2, adata3, join="outer")
    assert isinstance(adata_all.raw, Raw)
    assert set(adata_all.raw.var_names) == set("abcdz")
    assert set(adata_all.var_names) == set("abcd")
    assert not np.array_equal(
        np.nan_to_num(adata_all.raw.X), np.nan_to_num(adata_all.X)
    )

    del adata3.raw
    with pytest.warns(
        UserWarning,
        match=(
            "Only some AnnData objects have `.raw` attribute, "
            "not concatenating `.raw` attributes."
        ),
    ):
        adata_all = AnnData.concatenate(adata1, adata2, adata3)
    assert adata_all.raw is None

    del adata1.raw
    del adata2.raw
    assert all(_adata.raw is None for _adata in (adata1, adata2, adata3))
    adata_all = AnnData.concatenate(adata1, adata2, adata3)
    assert adata_all.raw is None


def test_concatenate_awkward(join_type):
    import awkward as ak

    a = ak.Array([[{"a": 1, "b": "foo"}], [{"a": 2, "b": "bar"}, {"a": 3, "b": "baz"}]])
    b = ak.Array([
        [{"a": 4}, {"a": 5}],
        [{"a": 6}],
        [{"a": 7}],
    ])

    adata_a = AnnData(np.zeros((2, 0), dtype=float), obsm={"awk": a})
    adata_b = AnnData(np.zeros((3, 0), dtype=float), obsm={"awk": b})

    if join_type == "inner":
        expected = ak.Array([
            [{"a": 1}],
            [{"a": 2}, {"a": 3}],
            [{"a": 4}, {"a": 5}],
            [{"a": 6}],
            [{"a": 7}],
        ])
    elif join_type == "outer":
        # TODO: This is what we would like to return, but waiting on:
        # * https://github.com/scikit-hep/awkward/issues/2182 and awkward 2.1.0
        # * https://github.com/scikit-hep/awkward/issues/2173
        # expected = ak.Array(
        #     [
        #         [{"a": 1, "b": "foo"}],
        #         [{"a": 2, "b": "bar"}, {"a": 3, "b": "baz"}],
        #         [{"a": 4, "b": None}, {"a": 5, "b": None}],
        #         [{"a": 6, "b": None}],
        #         [{"a": 7, "b": None}],
        #     ]
        # )
        expected = (
            ak.concatenate([  # I don't think I can construct a UnionArray directly
                ak.Array([
                    [{"a": 1, "b": "foo"}],
                    [{"a": 2, "b": "bar"}, {"a": 3, "b": "baz"}],
                ]),
                ak.Array([
                    [{"a": 4}, {"a": 5}],
                    [{"a": 6}],
                    [{"a": 7}],
                ]),
            ])
        )

    result = concat([adata_a, adata_b], join=join_type).obsm["awk"]

    assert_equal(expected, result)


@pytest.mark.parametrize(
    "other",
    [
        pd.DataFrame({"a": [4, 5, 6], "b": ["foo", "bar", "baz"]}, index=list("cde")),
        np.ones((3, 2)),
        sparse.random(3, 100, format="csr"),
    ],
)
def test_awkward_does_not_mix(join_type, other):
    import awkward as ak

    awk = ak.Array([
        [{"a": 1, "b": "foo"}],
        [{"a": 2, "b": "bar"}, {"a": 3, "b": "baz"}],
    ])

    adata_a = AnnData(
        np.zeros((2, 3), dtype=float),
        obs=pd.DataFrame(index=list("ab")),
        obsm={"val": awk},
    )
    adata_b = AnnData(
        np.zeros((3, 3), dtype=float),
        obs=pd.DataFrame(index=list("cde")),
        obsm={"val": other},
    )

    with pytest.raises(
        NotImplementedError,
        match=r"Cannot concatenate an AwkwardArray with other array types",
    ):
        concat([adata_a, adata_b], join=join_type)


def test_pairwise_concat(axis_name, array_type):
    axis, axis_name = merge._resolve_axis(axis_name)
    _, alt_axis_name = merge._resolve_axis(1 - axis)
    axis_sizes = [[100, 200, 50], [50, 50, 50]]
    if axis_name == "var":
        axis_sizes.reverse()
    Ms, Ns = axis_sizes
    axis_attr = f"{axis_name}p"
    alt_attr = f"{alt_axis_name}p"

    def gen_axis_array(m):
        return array_type(sparse.random(m, m, format="csr", density=0.1))

    adatas = {
        k: AnnData(
            X=sparse.csr_matrix((m, n)),
            obsp={"arr": gen_axis_array(m)},
            varp={"arr": gen_axis_array(n)},
        )
        for k, m, n in zip("abc", Ms, Ns, strict=True)
    }

    w_pairwise = concat(adatas, axis=axis, label="orig", pairwise=True)
    wo_pairwise = concat(adatas, axis=axis, label="orig", pairwise=False)

    # Check that argument controls whether elements are included
    assert getattr(wo_pairwise, axis_attr) == {}
    assert getattr(w_pairwise, axis_attr) != {}

    # Check values of included elements
    full_inds = np.arange(w_pairwise.shape[axis])
    obs_var: pd.DataFrame = getattr(w_pairwise, axis_name)
    groups = obs_var.groupby("orig", observed=True).indices
    for k, inds in groups.items():
        orig_arr = getattr(adatas[k], axis_attr)["arr"]
        full_arr = getattr(w_pairwise, axis_attr)["arr"]

        if isinstance(full_arr, DaskArray):
            full_arr = full_arr.compute()

        # Check original values are intact
        assert_equal(orig_arr, _subset(full_arr, (inds, inds)))
        # Check that entries are filled with zeroes
        assert_equal(
            sparse.csr_matrix((len(inds), len(full_inds) - len(inds))),
            _subset(full_arr, (inds, np.setdiff1d(full_inds, inds))),
        )
        assert_equal(
            sparse.csr_matrix((len(full_inds) - len(inds), len(inds))),
            _subset(full_arr, (np.setdiff1d(full_inds, inds), inds)),
        )

    # Check that argument does not affect alternative axis
    assert "arr" in getattr(
        concat(adatas, axis=axis, pairwise=False, merge="first"), alt_attr
    )


def test_nan_merge(axis_name, join_type, array_type):
    axis, _ = merge._resolve_axis(axis_name)
    alt_axis, alt_axis_name = merge._resolve_axis(1 - axis)
    mapping_attr = f"{alt_axis_name}m"
    adata_shape = (20, 10)
    # TODO: Revert to https://github.com/scverse/anndata/blob/71fdf821919fc5ff3c864dc74c4432c370573984/tests/test_concatenate.py#L961-L970 after https://github.com/scipy/scipy/pull/23626.
    # The need for this handling arose as a result of
    # https://github.com/dask/dask/pull/11755/files#diff-65211e64fa680da306e9612b92c60f557365507d46486325f0e7e04359bce64fR456-R459
    sparse_arr = sparse.random(adata_shape[alt_axis], 10, density=0.1, format="csr")
    sparse_arr_nan = sparse_arr.copy()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=sparse.SparseEfficiencyWarning)
        for _ in range(10):
            sparse_arr_nan[
                np.random.choice(sparse_arr.shape[0]),
                np.random.choice(sparse_arr.shape[1]),
            ] = np.nan
    arr = array_type(sparse_arr)
    arr_nan = array_type(sparse_arr_nan)

    _data = {"X": sparse.csr_matrix(adata_shape), mapping_attr: {"arr": arr_nan}}
    orig1 = AnnData(**_data)
    orig2 = AnnData(**_data)
    result = concat([orig1, orig2], axis=axis, join=join_type, merge="same")

    assert_equal(getattr(orig1, mapping_attr), getattr(result, mapping_attr))

    orig_nonan = AnnData(**{
        "X": sparse.csr_matrix(adata_shape),
        mapping_attr: {"arr": arr},
    })
    result_nonan = concat([orig1, orig_nonan], axis=axis, merge="same")

    assert len(getattr(result_nonan, mapping_attr)) == 0


def test_merge_unique():
    from anndata._core.merge import merge_unique

    # Simple cases
    assert merge_unique([{"a": "b"}, {"a": "b"}]) == {"a": "b"}
    assert merge_unique([{"a": {"b": "c"}}, {"a": {"b": "c"}}]) == {"a": {"b": "c"}}
    assert merge_unique([{"a": {"b": "c"}}, {"a": {"b": "d"}}]) == {}
    assert merge_unique([{"a": {"b": "c", "d": "e"}}, {"a": {"b": "c", "d": "f"}}]) == {
        "a": {"b": "c"}
    }

    assert merge_unique([
        {"a": {"b": {"c": {"d": "e"}}}},
        {"a": {"b": {"c": {"d": "e"}}}},
    ]) == {"a": {"b": {"c": {"d": "e"}}}}
    assert (
        merge_unique([
            {"a": {"b": {"c": {"d": "e"}}}},
            {"a": {"b": {"c": {"d": "f"}}}},
            {"a": {"b": {"c": {"d": "e"}}}},
        ])
        == {}
    )

    assert merge_unique([{"a": 1}, {"b": 2}]) == {"a": 1, "b": 2}
    assert merge_unique([{"a": 1}, {"b": 2}, {"a": 1, "b": {"c": 2, "d": 3}}]) == {
        "a": 1
    }

    # Test equivalency between arrays and lists
    assert list(
        merge_unique([{"a": np.ones(5)}, {"a": list(np.ones(5))}])["a"]
    ) == list(np.ones(5))
    assert merge_unique([{"a": np.ones(5)}, {"a": list(np.ones(4))}]) == {}


def test_merge_same():
    from anndata._core.merge import merge_same

    # Same as unique for a number of cases:
    assert merge_same([{"a": "b"}, {"a": "b"}]) == {"a": "b"}
    assert merge_same([{"a": {"b": "c"}}, {"a": {"b": "c"}}]) == {"a": {"b": "c"}}
    assert merge_same([{"a": {"b": "c"}}, {"a": {"b": "d"}}]) == {}
    assert merge_same([{"a": {"b": "c", "d": "e"}}, {"a": {"b": "c", "d": "f"}}]) == {
        "a": {"b": "c"}
    }

    assert merge_same([{"a": {"b": "c"}, "d": "e"}, {"a": {"b": "c"}, "d": 2}]) == {
        "a": {"b": "c"}
    }
    assert merge_same([
        {"a": {"b": {"c": {"d": "e"}}}},
        {"a": {"b": {"c": {"d": "e"}}}},
    ]) == {"a": {"b": {"c": {"d": "e"}}}}

    assert merge_same([{"a": 1}, {"b": 2}]) == {}
    assert merge_same([{"a": 1}, {"b": 2}, {"a": 1, "b": {"c": 2, "d": 3}}]) == {}

    # Test equivalency between arrays and lists
    assert list(merge_same([{"a": np.ones(5)}, {"a": list(np.ones(5))}])["a"]) == list(
        np.ones(5)
    )


def test_merge_first():
    from anndata._core.merge import merge_first

    assert merge_first([{"a": "b"}, {"a": "b"}]) == {"a": "b"}
    assert merge_first([{"a": {"b": "c"}}, {"a": {"b": "c"}}]) == {"a": {"b": "c"}}
    assert merge_first([{"a": 1}, {"a": 2}]) == {"a": 1}

    assert merge_first([{"a": 1}, {"a": {"b": {"c": {"d": "e"}}}}]) == {"a": 1}
    assert merge_first([{"a": {"b": {"c": {"d": "e"}}}}, {"a": 1}]) == {
        "a": {"b": {"c": {"d": "e"}}}
    }


# Helpers for test_concatenate_uns


def uns_ad(uns):
    return AnnData(np.zeros((10, 10)), uns=uns)


def map_values(mapping, path, key, old_parent, new_parent, new_items):
    ret = default_exit(path, key, old_parent, new_parent, new_items)
    for k, v in ret.items():
        if isinstance(v, Hashable) and v in mapping:
            ret[k] = mapping[v]
    return ret


def permute_nested_values(dicts: list[dict], gen_val: Callable[[int], Any]):
    """
    This function permutes the values of a nested mapping, for testing that out merge
    method work regardless of the values types.

    Assumes the initial dictionary had integers for values.
    """
    dicts = deepcopy(dicts)
    initial_values = [
        x[1] for x in research(dicts, query=lambda p, k, v: isinstance(v, int))
    ]
    mapping = {k: gen_val(k) for k in initial_values}
    return [remap(d, exit=partial(map_values, mapping)) for d in dicts]


def gen_df(n):
    return helpers.gen_typed_df(n)


def gen_array(n):
    return np.random.randn(n)


def gen_list(n):
    return list(gen_array(n))


def gen_sparse(n):
    return sparse.random(
        np.random.randint(1, 100), np.random.randint(1, 100), format="csr"
    )


def gen_something(n):
    options = [gen_df, gen_array, gen_list, gen_sparse]
    return np.random.choice(options)(n)


def gen_3d_numeric_array(n):
    return np.random.randn(n, n, n)


def gen_3d_recarray(_):
    # Ignoring n as it can get quite slow
    return gen_vstr_recarray(8, 3).reshape(2, 2, 2)


def gen_concat_params(unss, compat2result):
    value_generators = [
        lambda x: x,
        gen_df,
        gen_array,
        gen_list,
        gen_sparse,
        gen_something,
        gen_3d_numeric_array,
        gen_3d_recarray,
    ]
    for gen, (mode, result) in product(value_generators, compat2result.items()):
        yield pytest.param(unss, mode, result, gen)


@pytest.mark.parametrize(
    ("unss", "merge_strategy", "result", "value_gen"),
    chain(
        gen_concat_params(
            [{"a": 1}, {"a": 2}],
            {None: {}, "first": {"a": 1}, "unique": {}, "same": {}, "only": {}},
        ),
        gen_concat_params(
            [{"a": 1}, {"b": 2}],
            {
                None: {},
                "first": {"a": 1, "b": 2},
                "unique": {"a": 1, "b": 2},
                "same": {},
                "only": {"a": 1, "b": 2},
            },
        ),
        gen_concat_params(
            [
                {"a": {"b": 1, "c": {"d": 3}}},
                {"a": {"b": 1, "c": {"e": 4}}},
            ],
            {
                None: {},
                "first": {"a": {"b": 1, "c": {"d": 3, "e": 4}}},
                "unique": {"a": {"b": 1, "c": {"d": 3, "e": 4}}},
                "same": {"a": {"b": 1}},
                "only": {"a": {"c": {"d": 3, "e": 4}}},
            },
        ),
        gen_concat_params(
            [
                {"a": 1},
                {"a": 1, "b": 2},
                {"a": 1, "b": {"b.a": 1}, "c": 3},
                {"d": 4},
            ],
            {
                None: {},
                "first": {"a": 1, "b": 2, "c": 3, "d": 4},
                "unique": {"a": 1, "c": 3, "d": 4},
                "same": {},
                "only": {"c": 3, "d": 4},
            },
        ),
        gen_concat_params(
            [{"a": i} for i in range(15)],
            {None: {}, "first": {"a": 0}, "unique": {}, "same": {}, "only": {}},
        ),
        gen_concat_params(
            [{"a": 1} for i in range(10)] + [{"a": 2}],
            {None: {}, "first": {"a": 1}, "unique": {}, "same": {}, "only": {}},
        ),
    ),
)
def test_concatenate_uns(unss, merge_strategy, result, value_gen):
    """
    Test that concatenation works out for different strategies and sets of values.

    Params
    ------
    unss
        Set of patterns for values in uns.
    compat
        Strategy to use for merging uns.
    result
        Pattern we expect to see for the given input and strategy.
    value_gen
        Maps values in unss and results to another set of values. This is for checking that
        we're comparing values correctly. For example `[{"a": 1}, {"a": 1}]` may get mapped
        to `[{"a": [1, 2, 3]}, {"a": [1, 2, 3]}]`.
    """
    # So we can see what the initial pattern was meant to be
    print(merge_strategy, "\n", unss, "\n", result)
    result, *unss = permute_nested_values([result, *unss], value_gen)
    adatas = [uns_ad(uns) for uns in unss]
    with pytest.warns(FutureWarning, match=r"concatenate is deprecated"):
        merged = AnnData.concatenate(*adatas, uns_merge=merge_strategy).uns
    assert_equal(merged, result, elem_name="uns")


def test_transposed_concat(
    array_type,
    axis_name,
    join_type,
    merge_strategy,
    use_xdataset,
    force_lazy,
):
    axis, axis_name = merge._resolve_axis(axis_name)
    alt_axis = 1 - axis
    lhs = gen_adata(
        (10, 10),
        X_type=array_type,
        obs_xdataset=use_xdataset,
        var_xdataset=use_xdataset,
        **GEN_ADATA_DASK_ARGS,
    )
    rhs = gen_adata(
        (10, 12),
        X_type=array_type,
        obs_xdataset=use_xdataset,
        var_xdataset=use_xdataset,
        **GEN_ADATA_DASK_ARGS,
    )

    a = concat(
        [lhs, rhs],
        axis=axis,
        join=join_type,
        merge=merge_strategy,
        force_lazy=force_lazy,
    )
    b = concat(
        [lhs.T, rhs.T],
        axis=alt_axis,
        join=join_type,
        merge=merge_strategy,
        force_lazy=force_lazy,
    ).T

    assert_equal(a, b)


def test_batch_key(axis_name, use_xdataset, force_lazy):
    """Test that concat only adds a label if the key is provided"""

    get_annot = attrgetter(axis_name)

    lhs = gen_adata(
        (10, 10),
        obs_xdataset=use_xdataset,
        var_xdataset=use_xdataset,
        **GEN_ADATA_DASK_ARGS,
    )
    rhs = gen_adata(
        (10, 12),
        obs_xdataset=use_xdataset,
        var_xdataset=use_xdataset,
        **GEN_ADATA_DASK_ARGS,
    )

    # There is probably a prettier way to do this
    annot = get_annot(concat([lhs, rhs], axis=axis_name, force_lazy=force_lazy))
    assert (
        list(
            annot.columns.difference(
                get_annot(lhs).columns.union(get_annot(rhs).columns)
            )
        )
        == []
    )

    batch_annot = get_annot(
        concat([lhs, rhs], axis=axis_name, label="batch", force_lazy=force_lazy)
    )
    assert list(
        batch_annot.columns.difference(
            get_annot(lhs).columns.union(get_annot(rhs).columns)
        )
    ) == ["batch"]


def test_concat_categories_from_mapping(use_xdataset, force_lazy):
    mapping = {
        "a": gen_adata((10, 10), obs_xdataset=use_xdataset, var_xdataset=use_xdataset),
        "b": gen_adata((10, 10), obs_xdataset=use_xdataset, var_xdataset=use_xdataset),
    }
    keys = list(mapping.keys())
    adatas = list(mapping.values())

    mapping_call = partial(concat, mapping, force_lazy=force_lazy)
    iter_call = partial(concat, adatas, keys=keys, force_lazy=force_lazy)

    assert_equal(mapping_call(), iter_call())
    assert_equal(mapping_call(label="batch"), iter_call(label="batch"))
    assert_equal(mapping_call(index_unique="-"), iter_call(index_unique="-"))
    assert_equal(
        mapping_call(label="group", index_unique="+"),
        iter_call(label="group", index_unique="+"),
    )


def test_concat_categories_maintain_dtype():
    a = AnnData(
        X=np.ones((5, 1)),
        obs=pd.DataFrame(
            {
                "cat": pd.Categorical(list("aabcc")),
                "cat_ordered": pd.Categorical(list("aabcc"), ordered=True),
            },
            index=[f"cell{i:02}" for i in range(5)],
        ),
    )
    b = AnnData(
        X=np.ones((5, 1)),
        obs=pd.DataFrame(
            {
                "cat": pd.Categorical(list("bccdd")),
                "cat_ordered": pd.Categorical(list("bccdd"), ordered=True),
            },
            index=[f"cell{i:02}" for i in range(5, 10)],
        ),
    )
    c = AnnData(
        X=np.ones((5, 1)),
        obs=pd.DataFrame(
            {
                "cat_ordered": pd.Categorical(list("bccdd"), ordered=True),
            },
            index=[f"cell{i:02}" for i in range(5, 10)],
        ),
    )

    result = concat({"a": a, "b": b, "c": c}, join="outer")

    assert isinstance(result.obs["cat"].dtype, pd.CategoricalDtype), (
        f"Was {result.obs['cat'].dtype}"
    )
    assert pd.api.types.is_string_dtype(result.obs["cat_ordered"])


def test_concat_ordered_categoricals_retained():
    a = AnnData(
        X=np.ones((5, 1)),
        obs=pd.DataFrame(
            {
                "cat_ordered": pd.Categorical(list("aabcd"), ordered=True),
            },
            index=[f"cell{i:02}" for i in range(5)],
        ),
    )
    b = AnnData(
        X=np.ones((5, 1)),
        obs=pd.DataFrame(
            {
                "cat_ordered": pd.Categorical(list("abcdd"), ordered=True),
            },
            index=[f"cell{i:02}" for i in range(5, 10)],
        ),
    )

    c = concat([a, b])

    assert isinstance(c.obs["cat_ordered"].dtype, pd.CategoricalDtype)
    assert c.obs["cat_ordered"].cat.ordered


def test_concat_categorical_dtype_promotion():
    """https://github.com/scverse/anndata/issues/1170

    When concatenating categorical with other dtype, defer to pandas.
    """
    a = AnnData(
        np.ones((3, 3)),
        obs=pd.DataFrame(
            {"col": pd.Categorical(["a", "a", "b"])},
            index=[f"cell_{i:02d}" for i in range(3)],
        ),
    )
    b = AnnData(
        np.ones((3, 3)),
        obs=pd.DataFrame(
            {"col": ["c", "c", "c"]},
            index=[f"cell_{i:02d}" for i in range(3, 6)],
        ),
    )

    result = concat([a, b])
    expected = pd.concat([a.obs, b.obs])

    assert_equal(result.obs, expected)


def test_bool_promotion():
    np_bool = AnnData(
        np.ones((5, 1)),
        obs=pd.DataFrame({"bool": [True] * 5}, index=[f"cell{i:02}" for i in range(5)]),
    )
    missing = AnnData(
        np.ones((5, 1)),
        obs=pd.DataFrame(index=[f"cell{i:02}" for i in range(5, 10)]),
    )
    result = concat({"np_bool": np_bool, "b": missing}, join="outer", label="batch")

    assert pd.api.types.is_bool_dtype(result.obs["bool"])
    assert pd.isnull(result.obs.loc[result.obs["batch"] == "missing", "bool"]).all()

    # Check that promotion doesn't occur if it doesn't need to:
    np_bool_2 = AnnData(
        np.ones((5, 1)),
        obs=pd.DataFrame(
            {"bool": [True] * 5}, index=[f"cell{i:02}" for i in range(5, 10)]
        ),
    )
    result = concat(
        {"np_bool": np_bool, "np_bool_2": np_bool_2}, join="outer", label="batch"
    )

    assert result.obs["bool"].dtype == np.dtype(bool)


@pytest.mark.parametrize(
    ("index_unique", "expect_unique"),
    [
        pytest.param(None, False, id="default"),
        pytest.param("-", True, id="force_unique"),
    ],
)
@pytest.mark.parametrize("with_missing", [True, False], ids=["missing", "no_missing"])
def test_concat_names(
    *,
    axis_name: Literal["obs", "var"],
    use_xdataset: bool,
    force_lazy: bool,
    index_unique: str | None,
    expect_unique: bool,
    with_missing: bool,
) -> None:
    get_annot: attrgetter[pd.DataFrame] = attrgetter(axis_name)

    lhs = gen_adata((10, 10), obs_xdataset=use_xdataset, var_xdataset=use_xdataset)
    rhs = gen_adata((10, 10), obs_xdataset=use_xdataset, var_xdataset=use_xdataset)
    if with_missing:
        for s in [lhs, rhs]:
            new_index = pd.Index([*get_annot(s).index[:5], *([pd.NA] * 5)])
            setattr(s, f"{axis_name}_names", new_index)

    cat = concat(
        [lhs, rhs], axis=axis_name, index_unique=index_unique, force_lazy=force_lazy
    )
    assert get_annot(cat).index[~get_annot(cat).index.isna()].is_unique is expect_unique
    if with_missing:
        assert get_annot(cat).index.isna().sum() == 10


def axis_labels(adata: AnnData, axis: Literal[0, 1]) -> pd.Index:
    return (adata.obs_names, adata.var_names)[axis]


def expected_shape(
    a: AnnData, b: AnnData, axis: Literal[0, 1], join: Literal["inner", "outer"]
) -> tuple[int, int]:
    alt_axis = 1 - axis
    labels = partial(axis_labels, axis=alt_axis)
    shape = [None, None]

    shape[axis] = a.shape[axis] + b.shape[axis]
    if join == "inner":
        shape[alt_axis] = len(labels(a).intersection(labels(b)))
    elif join == "outer":
        shape[alt_axis] = len(labels(a).union(labels(b)))
    else:
        raise ValueError()

    return tuple(shape)


@pytest.mark.parametrize(
    "shape", [pytest.param((8, 0), id="no_var"), pytest.param((0, 10), id="no_obs")]
)
def test_concat_size_0_axis(
    axis_name, join_type, merge_strategy, shape, use_xdataset, force_lazy
):
    """Regression test for https://github.com/scverse/anndata/issues/526"""
    axis, axis_name = merge._resolve_axis(axis_name)
    alt_axis = 1 - axis
    col_dtypes = (*DEFAULT_COL_TYPES, pd.StringDtype())
    a = gen_adata(
        (5, 7),
        obs_dtypes=col_dtypes,
        var_dtypes=col_dtypes,
        obs_xdataset=use_xdataset,
        var_xdataset=use_xdataset,
    )
    b = gen_adata(
        shape,
        obs_dtypes=col_dtypes,
        var_dtypes=col_dtypes,
        obs_xdataset=use_xdataset,
        var_xdataset=use_xdataset,
    )

    expected_size = expected_shape(a, b, axis=axis, join=join_type)

    result = concat(
        {"a": a, "b": b},
        axis=axis,
        join=join_type,
        merge=merge_strategy,
        pairwise=True,
        index_unique="-",
        force_lazy=force_lazy,
    )
    assert result.shape == expected_size

    if join_type == "outer":
        # Check new entries along axis of concatenation
        axis_new_inds = axis_labels(result, axis).str.endswith("-b")
        altaxis_new_inds = ~axis_labels(result, alt_axis).isin(axis_labels(a, alt_axis))
        axis_idx = make_idx_tuple(axis_new_inds, axis)
        altaxis_idx = make_idx_tuple(altaxis_new_inds, 1 - axis)

        check_filled_like(result.X[axis_idx], elem_name="X")
        check_filled_like(result.X[altaxis_idx], elem_name="X")
        for k, elem in result.layers.items():
            check_filled_like(elem[axis_idx], elem_name=f"layers/{k}")
            check_filled_like(elem[altaxis_idx], elem_name=f"layers/{k}")

        if shape[axis] > 0:
            b_result = result[axis_idx].copy()
            mapping_elem = f"{axis_name}m"
            setattr(b_result, f"{axis_name}_names", getattr(b, f"{axis_name}_names"))
            for k, result_elem in getattr(b_result, mapping_elem).items():
                elem_name = f"{mapping_elem}/{k}"
                # pd.concat can have unintuitive return types. is similar to numpy promotion
                if isinstance(result_elem, pd.DataFrame):
                    assert_equal(
                        getattr(b, mapping_elem)[k].astype(object),
                        result_elem.astype(object),
                        elem_name=elem_name,
                    )
                else:
                    assert_equal(
                        getattr(b, mapping_elem)[k],
                        result_elem,
                        elem_name=elem_name,
                    )


@pytest.mark.parametrize("elem", ["sparse", "array", "df", "da"])
@pytest.mark.parametrize("axis", ["obs", "var"])
def test_concat_outer_aligned_mapping(elem, axis, use_xdataset, force_lazy):
    a = gen_adata(
        (5, 5),
        obs_xdataset=use_xdataset,
        var_xdataset=use_xdataset,
        **GEN_ADATA_DASK_ARGS,
    )
    b = gen_adata(
        (3, 5),
        obs_xdataset=use_xdataset,
        var_xdataset=use_xdataset,
        **GEN_ADATA_DASK_ARGS,
    )
    del getattr(b, f"{axis}m")[elem]

    concated = concat(
        {"a": a, "b": b}, join="outer", label="group", axis=axis, force_lazy=force_lazy
    )

    mask = getattr(concated, axis)["group"] == "b"
    result = getattr(
        concated[(mask, slice(None)) if axis == "obs" else (slice(None), mask)],
        f"{axis}m",
    )[elem]

    check_filled_like(result, elem_name=f"{axis}m/{elem}")


@mark_legacy_concatenate
def test_concatenate_size_0_axis():
    # https://github.com/scverse/anndata/issues/526

    a = gen_adata((5, 10))
    b = gen_adata((5, 0))

    # Mostly testing that this doesn't error
    assert a.concatenate([b]).shape == (10, 0)
    assert b.concatenate([a]).shape == (10, 0)


def test_concat_null_X(use_xdataset):
    adatas_orig = {
        k: gen_adata((20, 10), obs_xdataset=use_xdataset, var_xdataset=use_xdataset)
        for k in list("abc")
    }
    adatas_no_X = {}
    for k, v in adatas_orig.items():
        v = v.copy()
        del v.X
        adatas_no_X[k] = v

    orig = concat(adatas_orig, index_unique="-")
    no_X = concat(adatas_no_X, index_unique="-")
    del orig.X

    assert_equal(no_X, orig)


# https://github.com/scverse/ehrapy/issues/151#issuecomment-1016753744
@pytest.mark.parametrize("sparse_indexer_type", [np.int64, np.int32])
def test_concat_X_dtype(cpu_array_type, sparse_indexer_type):
    adatas_orig = {
        k: AnnData(cpu_array_type(np.ones((20, 10), dtype=np.int8)))
        for k in list("abc")
    }
    for adata in adatas_orig.values():
        adata.raw = AnnData(cpu_array_type(np.ones((20, 30), dtype=np.float64)))
        if sparse.issparse(adata.X):
            adata.X.indptr = adata.X.indptr.astype(sparse_indexer_type)
            adata.X.indices = adata.X.indices.astype(sparse_indexer_type)

    result = concat(adatas_orig, index_unique="-")

    assert result.X.dtype == np.int8
    assert result.raw.X.dtype == np.float64
    if sparse.issparse(result.X):
        # https://github.com/scipy/scipy/issues/20389 was merged in 1.15 but is still an issue with matrix
        if sparse_indexer_type == np.int64 and (
            (
                (issubclass(cpu_array_type, CSArray) or adata.X.format == "csc")
                and Version(version("scipy")) < Version("1.15.0")
            )
            or issubclass(cpu_array_type, CSMatrix)
        ):
            pytest.xfail(
                "Data type int64 is not maintained for sparse matrices or csc array"
            )
        assert result.X.indptr.dtype == sparse_indexer_type, result.X
        assert result.X.indices.dtype == sparse_indexer_type


# Leaving out for now. See definition of these values for explanation
# def test_concatenate_uns_types():
#     from anndata._core.merge import UNS_STRATEGIES, UNS_STRATEGIES_TYPE
#     assert set(UNS_STRATEGIES.keys()) == set(UNS_STRATEGIES_TYPE.__args__)


# Tests how dask plays with other types on concatenation.
def test_concat_different_types_dask(merge_strategy, array_type):
    import dask.array as da
    from scipy import sparse

    import anndata as ad

    varm_array = sparse.random(5, 20, density=0.5, format="csr")

    ad1 = ad.AnnData(X=np.ones((5, 5)), varm={"a": varm_array})
    ad1_other = ad.AnnData(X=np.ones((5, 5)), varm={"a": array_type(varm_array)})
    ad2 = ad.AnnData(X=np.zeros((5, 5)), varm={"a": da.ones(5, 20)})

    result1 = ad.concat([ad1, ad2], merge=merge_strategy)
    target1 = ad.concat([ad1_other, ad2], merge=merge_strategy)
    result2 = ad.concat([ad2, ad1], merge=merge_strategy)
    target2 = ad.concat([ad2, ad1_other], merge=merge_strategy)

    assert_equal(result1, target1)
    assert_equal(result2, target2)


def test_concat_missing_elem_dask_join(join_type):
    import dask.array as da

    import anndata as ad

    ad1 = ad.AnnData(X=np.ones((5, 10)))
    ad2 = ad.AnnData(X=np.zeros((5, 5)), layers={"a": da.ones((5, 5))})
    ad_in_memory_with_layers = ad2.to_memory()

    result1 = ad.concat([ad1, ad2], join=join_type)
    result2 = ad.concat([ad1, ad_in_memory_with_layers], join=join_type)
    assert_equal(result1, result2)


def test_impute_dask(axis_name):
    import dask.array as da

    from anndata._core.merge import _resolve_axis, missing_element

    axis, _ = _resolve_axis(axis_name)
    els = [da.ones((5, 5))]
    missing = missing_element(6, els, axis=axis, off_axis_size=17)
    assert isinstance(missing, DaskArray)
    in_memory = missing.compute()
    assert np.all(np.isnan(in_memory))
    assert in_memory.shape[axis] == 6
    assert in_memory.shape[axis - 1] == 17


def test_outer_concat_with_missing_value_for_df():
    # https://github.com/scverse/anndata/issues/901
    # TODO: Extend this test to cover all cases of missing values
    # TODO: Check values
    a_idx = ["a", "b", "c", "d", "e"]
    b_idx = ["f", "g", "h", "i", "j", "k", "l", "m"]
    a = AnnData(
        np.ones((5, 5)),
        obs=pd.DataFrame(index=a_idx),
    )
    b = AnnData(
        np.zeros((8, 9)),
        obs=pd.DataFrame(index=b_idx),
        obsm={"df": pd.DataFrame({"col": np.arange(8)}, index=b_idx)},
    )

    concat([a, b], join="outer")


def test_outer_concat_outputs_nullable_bool_writable(tmp_path):
    a = gen_adata((5, 5), obsm_types=(pd.DataFrame,))
    b = gen_adata((3, 5), obsm_types=(pd.DataFrame,))

    del b.obsm["df"]

    adatas = concat({"a": a, "b": b}, join="outer", label="group")
    adatas.write(tmp_path / "test.h5ad")


def test_concat_duplicated_columns(join_type):
    # https://github.com/scverse/anndata/issues/483
    a = AnnData(
        obs=pd.DataFrame(
            np.ones((5, 2)), columns=["a", "a"], index=[str(x) for x in range(5)]
        )
    )
    b = AnnData(
        obs=pd.DataFrame(
            np.ones((5, 1)), columns=["a"], index=[str(x) for x in range(5, 10)]
        )
    )

    with pytest.raises(pd.errors.InvalidIndexError, match=r"'a'"):
        concat([a, b], join=join_type)


@pytest.mark.gpu
def test_error_on_mixed_device():
    """https://github.com/scverse/anndata/issues/1083"""
    import cupy
    import cupyx.scipy.sparse as cupy_sparse

    cp_adata = AnnData(
        cupy.random.randn(10, 10),
        obs=pd.DataFrame(index=[f"cell_{i:02d}" for i in range(10)]),
    )
    cp_sparse_adata = AnnData(
        cupy_sparse.random(10, 10, format="csr", density=0.2),
        obs=pd.DataFrame(index=[f"cell_{i:02d}" for i in range(10, 20)]),
    )
    np_adata = AnnData(
        np.random.randn(10, 10),
        obs=pd.DataFrame(index=[f"cell_{i:02d}" for i in range(20, 30)]),
    )
    sparse_adata = AnnData(
        sparse.random(10, 10, format="csr", density=0.2),
        obs=pd.DataFrame(index=[f"cell_{i:02d}" for i in range(30, 40)]),
    )

    adatas = {
        "cupy": cp_adata,
        "cupy_sparse": cp_sparse_adata,
        "numpy": np_adata,
        "sparse": sparse_adata,
    }

    for p in map(dict, permutations(adatas.items())):
        print(list(p.keys()))
        with pytest.raises(
            NotImplementedError, match=r"Cannot concatenate a cupy array with other"
        ):
            concat(p)

    for p in permutations([cp_adata, cp_sparse_adata]):
        concat(p)


def test_concat_on_var_outer_join(array_type):
    # https://github.com/scverse/anndata/issues/1286
    a = AnnData(
        obs=pd.DataFrame(index=[f"cell_{i:02d}" for i in range(10)]),
        var=pd.DataFrame(index=[f"gene_{i:02d}" for i in range(10)]),
        layers={
            "X": array_type(np.ones((10, 10))),
        },
    )
    b = AnnData(
        obs=pd.DataFrame(index=[f"cell_{i:02d}" for i in range(10)]),
        var=pd.DataFrame(index=[f"gene_{i:02d}" for i in range(10, 20)]),
    )

    # This shouldn't error
    # TODO: specify expected result while accounting for null value
    _ = concat([a, b], join="outer", axis=1)


def test_concat_dask_sparse_matches_memory(join_type, merge_strategy):
    import dask.array as da

    X = sparse.random(50, 20, density=0.5, format="csr")
    X_dask = da.from_array(X, chunks=(5, 20))
    var_names_1 = [f"gene_{i}" for i in range(20)]
    var_names_2 = [f"gene_{i}{'_foo' if (i % 2) else ''}" for i in range(20)]

    ad1 = AnnData(X=X, var=pd.DataFrame(index=var_names_1))
    ad2 = AnnData(X=X, var=pd.DataFrame(index=var_names_2))

    ad1_dask = AnnData(X=X_dask, var=pd.DataFrame(index=var_names_1))
    ad2_dask = AnnData(X=X_dask, var=pd.DataFrame(index=var_names_2))

    res_in_memory = concat([ad1, ad2], join=join_type, merge=merge_strategy)
    res_dask = concat([ad1_dask, ad2_dask], join=join_type, merge=merge_strategy)
    assert_equal(res_in_memory, res_dask)


def test_1d_concat():
    adata = AnnData(np.ones((5, 20)), obsm={"1d-array": np.ones(5)})
    concated = concat([adata, adata])
    assert concated.obsm["1d-array"].shape == (10, 1)



from __future__ import annotations

from contextlib import contextmanager
from functools import partial, singledispatch
from pathlib import Path
from typing import TYPE_CHECKING, overload

import h5py
import numpy as np
import pandas as pd
from scipy import sparse

import anndata as ad
from anndata._core.file_backing import filename, get_elem_name
from anndata._core.xarray import Dataset2D, requires_xarray
from anndata.abc import CSCDataset, CSRDataset
from anndata.compat import (
    NULLABLE_NUMPY_STRING_TYPE,
    NUMPY_2,
    DaskArray,
    H5Array,
    H5Group,
    XDataset,
    XVariable,
    ZarrArray,
    ZarrGroup,
)

from .registry import _LAZY_REGISTRY, IOSpec, read_elem

if TYPE_CHECKING:
    from collections.abc import Generator, Mapping, Sequence
    from typing import Literal, ParamSpec, TypeVar

    from anndata.experimental.backed._lazy_arrays import CategoricalArray, MaskedArray

    from ...compat import CSArray, CSMatrix, H5File
    from .registry import LazyDataStructures, LazyReader

    BlockInfo = Mapping[
        None,
        dict[str, Sequence[tuple[int, int]]],
    ]

    P = ParamSpec("P")
    R = TypeVar("R")
    D = TypeVar("D")


@overload
@contextmanager
def maybe_open_h5(
    path_or_other: Path, elem_name: str
) -> Generator[H5File, None, None]: ...
@overload
@contextmanager
def maybe_open_h5(path_or_other: D, elem_name: str) -> Generator[D, None, None]: ...
@contextmanager
def maybe_open_h5(
    path_or_other: H5File | D, elem_name: str
) -> Generator[H5File | D, None, None]:
    if not isinstance(path_or_other, Path):
        yield path_or_other
        return
    file = h5py.File(path_or_other, "r")
    try:
        yield file[elem_name]
    finally:
        file.close()


_DEFAULT_STRIDE = 1000


def compute_chunk_layout_for_axis_size(
    chunk_axis_size: int, full_axis_size: int
) -> tuple[int, ...]:
    n_strides, rest = np.divmod(full_axis_size, chunk_axis_size)
    chunk = (chunk_axis_size,) * n_strides
    if rest > 0:
        chunk += (rest,)
    return chunk


def make_dask_chunk(
    path_or_sparse_dataset: Path | D,
    elem_name: str,
    block_info: BlockInfo | None = None,
) -> CSMatrix | CSArray:
    if block_info is None:
        msg = "Block info is required"
        raise ValueError(msg)
    # We need to open the file in each task since `dask` cannot share h5py objects when using `dask.distributed`
    # https://github.com/scverse/anndata/issues/1105
    with maybe_open_h5(path_or_sparse_dataset, elem_name) as f:
        # See https://github.com/scverse/anndata/pull/2005 for why
        # should_cache_indptr is False.
        # The prupose of caching the indptr was when the dataset is reused
        # which is in general the case but is not here.  Hence
        # caching it on every access to the dataset here is quite costly.
        mtx = (
            ad.io.sparse_dataset(f, should_cache_indptr=False)
            if isinstance(f, H5Group)
            else f
        )
        idx = tuple(
            slice(start, stop) for start, stop in block_info[None]["array-location"]
        )
        chunk = mtx[idx]
    return chunk


@singledispatch
def get_chunksize(obj) -> tuple[int, ...]:
    if hasattr(obj, "chunks"):
        return obj.chunks
    msg = "object of type {type(obj)} has no recognized chunks"
    raise ValueError(msg)


@_LAZY_REGISTRY.register_read(H5Group, IOSpec("csc_matrix", "0.1.0"))
@_LAZY_REGISTRY.register_read(H5Group, IOSpec("csr_matrix", "0.1.0"))
@_LAZY_REGISTRY.register_read(ZarrGroup, IOSpec("csc_matrix", "0.1.0"))
@_LAZY_REGISTRY.register_read(ZarrGroup, IOSpec("csr_matrix", "0.1.0"))
def read_sparse_as_dask(
    elem: H5Group | ZarrGroup,
    *,
    _reader: LazyReader,
    chunks: tuple[int, ...] | None = None,  # only tuple[int, int] is supported here
) -> DaskArray:
    import dask.array as da

    path_or_sparse_dataset = (
        Path(filename(elem))
        if isinstance(elem, H5Group)
        else ad.io.sparse_dataset(elem, should_cache_indptr=False)
    )
    elem_name = get_elem_name(elem)
    shape: tuple[int, int] = tuple(elem.attrs["shape"])
    if isinstance(path_or_sparse_dataset, CSRDataset | CSCDataset):
        dtype = path_or_sparse_dataset.dtype
    else:
        dtype = elem["data"].dtype
    is_csc: bool = elem.attrs["encoding-type"] == "csc_matrix"

    stride: int = _DEFAULT_STRIDE
    major_dim, minor_dim = (1, 0) if is_csc else (0, 1)
    if chunks is not None:
        if len(chunks) != 2:
            msg = "`chunks` must be a tuple of two integers"
            raise ValueError(msg)
        if chunks[minor_dim] not in {shape[minor_dim], -1, None}:
            msg = (
                "Only the major axis can be chunked. "
                f"Try setting chunks to {((-1, _DEFAULT_STRIDE) if is_csc else (_DEFAULT_STRIDE, -1))}"
            )
            raise ValueError(msg)
        stride = (
            chunks[major_dim]
            if chunks[major_dim] not in {None, -1}
            else shape[major_dim]
        )

    shape_minor, shape_major = shape if is_csc else shape[::-1]
    chunks_major = compute_chunk_layout_for_axis_size(stride, shape_major)
    chunks_minor = (shape_minor,)
    chunk_layout = (
        (chunks_minor, chunks_major) if is_csc else (chunks_major, chunks_minor)
    )
    memory_format = sparse.csc_matrix if is_csc else sparse.csr_matrix
    make_chunk = partial(make_dask_chunk, path_or_sparse_dataset, elem_name)
    da_mtx = da.map_blocks(
        make_chunk,
        dtype=dtype,
        chunks=chunk_layout,
        meta=memory_format((0, 0), dtype=dtype),
    )
    return da_mtx


def resolve_chunks(
    elem: H5Array | ZarrArray,
    chunks_arg: tuple[int, ...] | None,
    shape: tuple[int, ...],
) -> tuple[int, ...]:
    shape = tuple(elem.shape)
    if chunks_arg is not None:
        # None and -1 on a given axis indicate that one should use the shape
        # in `dask`'s semantics.
        return tuple(
            c if c not in {None, -1} else s
            for c, s in zip(chunks_arg, shape, strict=True)
        )
    elif elem.chunks is None:  # h5 unchunked
        return tuple(min(_DEFAULT_STRIDE, s) for s in shape)
    return elem.chunks


# TODO: `map_blocks` of a string array in h5py is so insanely slow on benchmarking that in the case someone has
# a pure string annotation (not categoricals! or nullables strings!), it's probably better to pay the memory penalty.
# In the long run, it might be good to figure out what exactly is going on here but for now, this will do.
@_LAZY_REGISTRY.register_read(H5Array, IOSpec("string-array", "0.2.0"))
def read_h5_string_array(
    elem: H5Array,
    *,
    _reader: LazyReader,
    chunks: tuple[int] | None = None,
) -> DaskArray:
    import dask.array as da

    chunks = resolve_chunks(elem, chunks, tuple(elem.shape))
    return da.from_array(read_elem(elem), chunks=chunks)


@_LAZY_REGISTRY.register_read(H5Array, IOSpec("array", "0.2.0"))
def read_h5_array(
    elem: H5Array, *, _reader: LazyReader, chunks: tuple[int, ...] | None = None
) -> DaskArray:
    import dask.array as da

    path = Path(elem.file.filename)
    elem_name: str = elem.name
    shape = tuple(elem.shape)
    dtype = elem.dtype
    chunks = resolve_chunks(elem, chunks, shape)

    chunk_layout = tuple(
        compute_chunk_layout_for_axis_size(chunks[i], shape[i])
        for i in range(len(shape))
    )

    make_chunk = partial(make_dask_chunk, path, elem_name)
    return da.map_blocks(
        make_chunk, dtype=dtype, chunks=chunk_layout, meta=np.array([])
    )


@_LAZY_REGISTRY.register_read(ZarrArray, IOSpec("string-array", "0.2.0"))
@_LAZY_REGISTRY.register_read(ZarrArray, IOSpec("array", "0.2.0"))
def read_zarr_array(
    elem: ZarrArray, *, _reader: LazyReader, chunks: tuple[int, ...] | None = None
) -> DaskArray:
    import dask.array as da

    return da.from_zarr(elem, chunks=chunks)


def _gen_xarray_dict_iterator_from_elems(
    elem_dict: dict[str, LazyDataStructures],
    dim_name: str,
    index: np.NDArray,
) -> Generator[tuple[str, XVariable], None, None]:
    from anndata.experimental.backed._lazy_arrays import CategoricalArray, MaskedArray

    from ...compat import xarray as xr

    for k, v in elem_dict.items():
        if isinstance(v, DaskArray) and k != dim_name:
            variable = xr.Variable([dim_name], data=v)
        elif isinstance(v, CategoricalArray | MaskedArray) and k != dim_name:
            variable = xr.Variable(
                [dim_name],
                data=xr.core.indexing.LazilyIndexedArray(v),
                attrs={
                    "base_path_or_zarr_group": v.base_path_or_zarr_group,
                    "elem_name": v.elem_name,
                    "is_nullable_string": isinstance(v, MaskedArray)
                    and (
                        v.dtype == NULLABLE_NUMPY_STRING_TYPE
                        or isinstance(v.dtype, pd.StringDtype | np.dtypes.StringDType)
                    ),
                },
            )
        elif k == dim_name:
            variable = xr.Variable([dim_name], data=index)
        else:
            msg = f"Could not read {k}: {v} from into xarray Dataset2D"
            raise ValueError(msg)
        yield k, variable


DUMMY_RANGE_INDEX_KEY = "_anndata_dummy_range_index"


@_LAZY_REGISTRY.register_read(ZarrGroup, IOSpec("dataframe", "0.2.0"))
@_LAZY_REGISTRY.register_read(H5Group, IOSpec("dataframe", "0.2.0"))
@requires_xarray
def read_dataframe(
    elem: H5Group | ZarrGroup,
    *,
    _reader: LazyReader,
    use_range_index: bool = False,
    chunks: tuple[int] | None = None,
) -> Dataset2D:
    from xarray.core.indexing import BasicIndexer

    from ...experimental.backed._lazy_arrays import MaskedArray

    elem_dict = {
        k: _reader.read_elem(elem[k], chunks=chunks)
        for k in [*elem.attrs["column-order"], elem.attrs["_index"]]
    }
    # If we use a range index, the coord axis needs to have the special dim name
    # which is used below as well.
    if not use_range_index:
        dim_name = elem.attrs["_index"]
        # no sense in reading this in multiple times since xarray requires an in-memory index
        if isinstance(elem_dict[dim_name], DaskArray):
            index = elem_dict[dim_name].compute()
        elif isinstance(elem_dict[dim_name], MaskedArray):
            index = elem_dict[dim_name][BasicIndexer((slice(None),))]
        else:
            raise NotImplementedError()
    else:
        dim_name = DUMMY_RANGE_INDEX_KEY
        index = pd.RangeIndex(len(elem_dict[elem.attrs["_index"]])).astype("str")
    elem_xarray_dict = dict(
        _gen_xarray_dict_iterator_from_elems(elem_dict, dim_name, index)
    )
    if use_range_index:
        elem_xarray_dict[DUMMY_RANGE_INDEX_KEY] = XVariable(
            [DUMMY_RANGE_INDEX_KEY],
            data=index,
        )
    ds = Dataset2D(XDataset(elem_xarray_dict))
    ds.is_backed = True
    # We ensure the indexing_key attr always points to the true index
    # so that the roundtrip works even for the `use_range_index` `True` case
    ds.true_index_dim = elem.attrs["_index"]
    return ds


@_LAZY_REGISTRY.register_read(ZarrGroup, IOSpec("categorical", "0.2.0"))
@_LAZY_REGISTRY.register_read(H5Group, IOSpec("categorical", "0.2.0"))
@requires_xarray
def read_categorical(
    elem: H5Group | ZarrGroup,
    *,
    _reader: LazyReader,
) -> CategoricalArray:
    from anndata.experimental.backed._lazy_arrays import CategoricalArray

    base_path_or_zarr_group = (
        Path(filename(elem)) if isinstance(elem, H5Group) else elem
    )
    elem_name = get_elem_name(elem)
    return CategoricalArray(
        codes=elem["codes"],
        categories=elem["categories"],
        ordered=elem.attrs["ordered"],
        base_path_or_zarr_group=base_path_or_zarr_group,
        elem_name=elem_name,
    )


@requires_xarray
def read_nullable(
    elem: H5Group | ZarrGroup,
    *,
    encoding_type: Literal[
        "nullable-integer", "nullable-boolean", "nullable-string-array"
    ],
    _reader: LazyReader,
) -> MaskedArray:
    from anndata.experimental.backed._lazy_arrays import MaskedArray

    base_path_or_zarr_group = (
        Path(filename(elem)) if isinstance(elem, H5Group) else elem
    )
    elem_name = get_elem_name(elem)
    values = elem["values"]
    # HDF5 stores strings as bytes; use .astype("T") to decode on access
    # h5py recommends .astype("T") over .asstr() when using numpy ≥2
    if encoding_type == "nullable-string-array" and isinstance(elem, H5Group):
        values = values.astype("T") if NUMPY_2 else values.asstr()
    return MaskedArray(
        values=values,
        mask=elem["mask"],
        dtype_str=encoding_type,
        base_path_or_zarr_group=base_path_or_zarr_group,
        elem_name=elem_name,
    )


for dtype in ["integer", "boolean", "string-array"]:
    for group_type in [ZarrGroup, H5Group]:
        _LAZY_REGISTRY.register_read(group_type, IOSpec(f"nullable-{dtype}", "0.1.0"))(
            partial(read_nullable, encoding_type=f"nullable-{dtype}")
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol, runtime_checkable

if TYPE_CHECKING:
    from ._core.anndata import AnnData


@runtime_checkable
class ExtensionNamespace(Protocol):
    """Protocol for extension namespaces.

    Enforces that the namespace initializer accepts a class with the proper `__init__` method.
    Protocol's can't enforce that the `__init__` accepts the correct types. See
    `_check_namespace_signature` for that. This is mainly useful for static type
    checking with mypy and IDEs.
    """

    def __init__(self, adata: AnnData) -> None:
        """
        Used to enforce the correct signature for extension namespaces.
        """



from __future__ import annotations

from pathlib import Path
from subprocess import run
from typing import TYPE_CHECKING

import pandas as pd
import pytest
import zarr
import zarr.storage
from scipy import sparse

import anndata as ad
from anndata.compat import is_zarr_v2
from anndata.tests.helpers import assert_equal

if TYPE_CHECKING:
    from typing import Literal

ARCHIVE_PTH = Path(__file__).parent / "data/archives"


@pytest.fixture(params=list(ARCHIVE_PTH.glob("v*")), ids=lambda x: x.name)
def archive_dir(request: pytest.FixtureRequest) -> Path:
    assert isinstance(request.param, Path)
    if request.param.name == "v0.5.0":
        reason = "v0.5.0 has no zarr and very different format"
        request.applymarker(pytest.mark.xfail(reason=reason))
    return request.param


def read_archive(
    archive_dir: Path, format: Literal["h5ad", "zarr"]
) -> tuple[ad.AnnData, Path]:
    if format == "h5ad":
        path = archive_dir / "adata.h5ad"
        return ad.read_h5ad(path), path
    if format == "zarr":
        path = archive_dir / "adata.zarr.zip"
        store = path if is_zarr_v2() else zarr.storage.ZipStore(path)
        return ad.read_zarr(store), path
    pytest.fail(f"Unknown format: {format}")


@pytest.mark.filterwarnings("ignore::anndata.OldFormatWarning")
def test_backwards_compat_files(archive_dir: Path) -> None:
    from_h5ad, _ = read_archive(archive_dir, "h5ad")
    from_zarr, _ = read_archive(archive_dir, "zarr")
    assert_equal(from_h5ad, from_zarr, exact=True)


def test_no_diff(tmp_path: Path, archive_dir: Path) -> None:
    if archive_dir.name in {"v0.7.8", "v0.7.0"}:
        pytest.skip("DataFrame encoding changed between 0.7 and now")
    adata, in_path = read_archive(archive_dir, "h5ad")
    with ad.settings.override(allow_write_nullable_strings=False):
        adata.write_h5ad(out_path := tmp_path / "adata.h5ad")
    diff_proc = run(
        ["h5diff", "-c", in_path, out_path], check=False, capture_output=True, text=True
    )
    assert diff_proc.returncode == 0, diff_proc.stdout


def test_clean_uns_backwards_compat(tmp_path, diskfmt):
    pth = tmp_path / f"test_write.{diskfmt}"
    write = lambda x, y: getattr(x, f"write_{diskfmt}")(y)
    read = lambda x: getattr(ad, f"read_{diskfmt}")(x)

    orig = ad.AnnData(
        sparse.csr_matrix((3, 5), dtype="float32"),
        obs=pd.DataFrame(
            {"a": pd.Categorical(list("aab")), "b": [1, 2, 3]},
            index=[f"cell_{i}" for i in range(3)],
        ),
        uns={
            "a_categories": "some string",
            "b_categories": "another string",
        },
    )

    write(orig, pth)
    from_disk = read(pth)

    assert_equal(orig, from_disk)



"""Annotated multivariate observation data."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any

from ._core.anndata import AnnData
from ._core.extensions import register_anndata_namespace
from ._core.merge import concat
from ._core.raw import Raw
from ._settings import settings
from ._warnings import (
    ExperimentalFeatureWarning,
    ImplicitModificationWarning,
    OldFormatWarning,
    WriteWarning,
)
from .io import read_h5ad, read_zarr
from .utils import module_get_attr_redirect

# Submodules need to be imported last
from . import abc, experimental, typing, io, types  # isort: skip

# We use these in tests by attribute access
from . import logging  # noqa: F401  # isort: skip

__all__ = [
    "AnnData",
    "ExperimentalFeatureWarning",
    "ImplicitModificationWarning",
    "OldFormatWarning",
    "Raw",
    "WriteWarning",
    "abc",
    "concat",
    "experimental",
    "io",
    "read_h5ad",
    "read_zarr",
    "register_anndata_namespace",
    "settings",
    "types",
    "typing",
]

_DEPRECATED_IO = (
    "read_loom",
    "read_hdf",
    "read_excel",
    "read_umi_tools",
    "read_csv",
    "read_text",
    "read_mtx",
)
_DEPRECATED = {method: f"io.{method}" for method in _DEPRECATED_IO}


def __getattr__(attr_name: str) -> Any:
    if attr_name == "__version__":
        import warnings
        from importlib.metadata import version

        msg = "`__version__` is deprecated, use `importlib.metadata.version('anndata')` instead."
        warnings.warn(msg, FutureWarning, stacklevel=2)
        return version("anndata")

    return module_get_attr_redirect(attr_name, deprecated_mapping=_DEPRECATED)



from __future__ import annotations

from collections.abc import Callable
from typing import TypeVar

F = TypeVar("F", bound=Callable)


def doctest_needs(mod: str) -> Callable[[F], F]:
    """Mark function with doctest dependency."""

    def decorator(func: F) -> F:
        func._doctest_needs = mod
        return func

    return decorator



from __future__ import annotations

import warnings

__all__: list[str] = []


def __getattr__(key: str):
    from .. import io

    attr = getattr(io, key)
    msg = (
        f"Importing {key} from `anndata._io` is deprecated. "
        "Please use anndata.io instead."
    )
    warnings.warn(msg, FutureWarning, stacklevel=2)
    return attr



"""\
This file contains tests for conversion made during io.
"""

from __future__ import annotations

import h5py
import numpy as np
import pytest
from scipy import sparse

import anndata as ad
from anndata.compat import CSMatrix
from anndata.tests.helpers import GEN_ADATA_NO_XARRAY_ARGS, assert_equal, gen_adata


@pytest.fixture(
    params=[sparse.csr_matrix, sparse.csc_matrix, np.array],
    ids=["scipy-csr", "scipy-csc", "np-array"],
)
def mtx_format(request):
    return request.param


@pytest.fixture(
    params=[sparse.csr_matrix, sparse.csc_matrix],
    ids=["scipy-csr", "scipy-csc"],
)
def spmtx_format(request):
    return request.param


@pytest.fixture(params=[("raw/X",), ("X",), ("X", "raw/X")])
def to_convert(request):
    return request.param


def test_sparse_to_dense_disk(tmp_path, mtx_format, to_convert):
    mem_pth = tmp_path / "orig.h5ad"
    dense_from_mem_pth = tmp_path / "dense_mem.h5ad"
    dense_from_disk_pth = tmp_path / "dense_disk.h5ad"
    mem = gen_adata((50, 50), mtx_format, **GEN_ADATA_NO_XARRAY_ARGS)
    mem.raw = mem.copy()

    mem.write_h5ad(mem_pth)
    disk = ad.read_h5ad(mem_pth, backed="r")

    mem.write_h5ad(dense_from_mem_pth, as_dense=to_convert)
    disk.write_h5ad(dense_from_disk_pth, as_dense=to_convert)

    with h5py.File(dense_from_mem_pth, "r") as f:
        for k in to_convert:
            assert isinstance(f[k], h5py.Dataset)
    with h5py.File(dense_from_disk_pth, "r") as f:
        for k in to_convert:
            assert isinstance(f[k], h5py.Dataset)

    for backed in [None, "r"]:
        from_mem = ad.read_h5ad(dense_from_mem_pth, backed=backed)
        from_disk = ad.read_h5ad(dense_from_disk_pth, backed=backed)
        assert_equal(mem, from_mem)
        assert_equal(mem, from_disk)
        assert_equal(disk, from_mem)
        assert_equal(disk, from_disk)


def test_sparse_to_dense_inplace(tmp_path, spmtx_format):
    pth = tmp_path / "adata.h5ad"
    orig = gen_adata((50, 50), spmtx_format, **GEN_ADATA_NO_XARRAY_ARGS)
    orig.raw = orig.copy()
    orig.write(pth)
    backed = ad.read_h5ad(pth, backed="r+")
    backed.write(as_dense=("X", "raw/X"))
    new = ad.read_h5ad(pth)

    assert_equal(orig, new)
    assert_equal(backed, new)

    assert isinstance(new.X, np.ndarray)
    assert isinstance(new.raw.X, np.ndarray)
    assert isinstance(orig.X, spmtx_format)
    assert isinstance(orig.raw.X, spmtx_format)
    assert isinstance(backed.X, h5py.Dataset)
    assert isinstance(backed.raw.X, h5py.Dataset)


def test_sparse_to_dense_errors(tmp_path):
    adata = ad.AnnData(X=sparse.random(50, 50, format="csr"))
    adata.layers["like_X"] = adata.X.copy()
    with pytest.raises(ValueError, match=r"Cannot specify writing"):
        adata.write_h5ad(tmp_path / "failure.h5ad", as_dense=("raw/X",))
    with pytest.raises(NotImplementedError):
        adata.write_h5ad(tmp_path / "failure.h5ad", as_dense=("raw", "X"))
    with pytest.raises(NotImplementedError):
        adata.write_h5ad(tmp_path / "failure.h5ad", as_dense=("layers/like_X",))


def test_dense_to_sparse_memory(tmp_path, spmtx_format, to_convert):
    dense_path = tmp_path / "dense.h5ad"
    orig = gen_adata((50, 50), np.array, **GEN_ADATA_NO_XARRAY_ARGS)
    orig.raw = orig.copy()
    orig.write_h5ad(dense_path)
    assert not isinstance(orig.X, CSMatrix)
    assert not isinstance(orig.raw.X, CSMatrix)

    curr = ad.read_h5ad(dense_path, as_sparse=to_convert, as_sparse_fmt=spmtx_format)

    if "X" in to_convert:
        assert isinstance(curr.X, spmtx_format)
    if "raw/X" in to_convert:
        assert isinstance(curr.raw.X, spmtx_format)

    assert_equal(orig, curr)


def test_dense_to_sparse_errors(tmp_path):
    dense_pth = tmp_path / "dense.h5ad"
    adata = ad.AnnData(X=np.ones((50, 50)))
    adata.layers["like_X"] = adata.X.copy()
    adata.write(dense_pth)

    with pytest.raises(NotImplementedError):
        ad.read_h5ad(dense_pth, as_sparse=("X",), as_sparse_fmt=sparse.coo_matrix)
    with pytest.raises(NotImplementedError):
        ad.read_h5ad(dense_pth, as_sparse=("layers/like_X",))



from __future__ import annotations

import inspect
from pathlib import Path
from typing import TYPE_CHECKING, Generic, TypeVar, get_type_hints, overload
from warnings import warn

from ..types import ExtensionNamespace
from .anndata import AnnData

if TYPE_CHECKING:
    from collections.abc import Callable


# Based off of the extension framework in Polars
# https://github.com/pola-rs/polars/blob/main/py-polars/polars/api.py

__all__ = ["register_anndata_namespace"]


def find_stacklevel() -> int:
    """
    Find the first place in the stack that is not inside AnnData.

    Taken from:
    https://github.com/pola-rs/polars/blob/main/py-polars/polars/_utils/various.py#L447
    """

    pkg_dir = str(Path(__file__).parent.parent)

    # https://stackoverflow.com/questions/17407119/python-inspect-stack-is-slow
    frame = inspect.currentframe()
    n = 0
    try:
        while frame:
            fname = inspect.getfile(frame)
            if fname.startswith(pkg_dir) or (
                (qualname := getattr(frame.f_code, "co_qualname", None))
                # ignore @singledispatch wrappers
                and qualname.startswith("singledispatch.")
            ):
                frame = frame.f_back
                n += 1
            else:
                break
    finally:
        # https://docs.python.org/3/library/inspect.html
        # > Though the cycle detector will catch these, destruction of the frames
        # > (and local variables) can be made deterministic by removing the cycle
        # > in a finally clause.
        del frame
    return n


# Reserved namespaces include accessors built into AnnData (currently there are none)
# and all current attributes of AnnData
_reserved_namespaces: set[str] = set(dir(AnnData))

NameSpT = TypeVar("NameSpT", bound=ExtensionNamespace)
T = TypeVar("T")


class AccessorNameSpace(ExtensionNamespace, Generic[NameSpT]):
    """Establish property-like namespace object for user-defined functionality."""

    def __init__(self, name: str, namespace: type[NameSpT]) -> None:
        self._accessor = name
        self._ns = namespace

    @overload
    def __get__(self, instance: None, cls: type[T]) -> type[NameSpT]: ...

    @overload
    def __get__(self, instance: T, cls: type[T]) -> NameSpT: ...

    def __get__(self, instance: T | None, cls: type[T]) -> NameSpT | type[NameSpT]:
        if instance is None:
            return self._ns

        ns_instance = self._ns(instance)  # type: ignore[call-arg]
        setattr(instance, self._accessor, ns_instance)
        return ns_instance


def _check_namespace_signature(ns_class: type) -> None:
    """Validate the signature of a namespace class for AnnData extensions.

    This function ensures that any class intended to be used as an extension namespace
    has a properly formatted `__init__` method such that:

    1. Accepts at least two parameters (self and adata)
    2. Has 'adata' as the name of the second parameter
    3. Has the second parameter properly type-annotated as 'AnnData' or any equivalent import alias

    The function performs runtime validation of these requirements before a namespace
    can be registered through the `register_anndata_namespace` decorator.

    Parameters
    ----------
    ns_class
        The namespace class to validate.

    Raises
    ------
    TypeError
        If the `__init__` method has fewer than 2 parameters (missing the AnnData parameter).
    AttributeError
        If the second parameter of `__init__` lacks a type annotation.
    TypeError
        If the second parameter of `__init__` is not named 'adata'.
    TypeError
        If the second parameter of `__init__` is not annotated as the 'AnnData' class.
    TypeError
        If both the name and type annotation of the second parameter are incorrect.

    """
    sig = inspect.signature(ns_class.__init__)
    params = list(sig.parameters.values())

    # Ensure there are at least two parameters (self and adata)
    if len(params) < 2:
        error_msg = "Namespace initializer must accept an AnnData instance as the second parameter."
        raise TypeError(error_msg)

    # Get the second parameter (expected to be 'adata')
    param = params[1]
    if param.annotation is inspect._empty:
        err_msg = "Namespace initializer's second parameter must be annotated as the 'AnnData' class, got empty annotation."
        raise AttributeError(err_msg)

    name_ok = param.name == "adata"

    # Resolve the annotation using get_type_hints to handle forward references and aliases.
    try:
        type_hints = get_type_hints(ns_class.__init__)
        resolved_type = type_hints.get(param.name, param.annotation)
    except NameError as e:
        err_msg = f"Namespace initializer's second parameter must be named 'adata', got '{param.name}'."
        raise NameError(err_msg) from e

    type_ok = resolved_type is AnnData

    match (name_ok, type_ok):
        case (True, True):
            return  # Signature is correct.
        case (False, True):
            msg = f"Namespace initializer's second parameter must be named 'adata', got {param.name!r}."
            raise TypeError(msg)
        case (True, False):
            type_repr = getattr(resolved_type, "__name__", str(resolved_type))
            msg = f"Namespace initializer's second parameter must be annotated as the 'AnnData' class, got '{type_repr}'."
            raise TypeError(msg)
        case _:
            type_repr = getattr(resolved_type, "__name__", str(resolved_type))
            msg = (
                f"Namespace initializer's second parameter must be named 'adata', got {param.name!r}. "
                f"And must be annotated as 'AnnData', got {type_repr!r}."
            )
            raise TypeError(msg)


def _create_namespace(
    name: str, cls: type[AnnData]
) -> Callable[[type[NameSpT]], type[NameSpT]]:
    """Register custom namespace against the underlying AnnData class."""

    def namespace(ns_class: type[NameSpT]) -> type[NameSpT]:
        _check_namespace_signature(ns_class)  # Perform the runtime signature check
        if name in _reserved_namespaces:
            msg = f"cannot override reserved attribute {name!r}"
            raise AttributeError(msg)
        elif name in cls._accessors:
            warn(
                f"Overriding existing custom namespace {name!r} (on {cls.__name__!r})",
                UserWarning,
                stacklevel=find_stacklevel(),
            )
        setattr(cls, name, AccessorNameSpace(name, ns_class))
        cls._accessors.add(name)
        return ns_class

    return namespace


def register_anndata_namespace(
    name: str,
) -> Callable[[type[NameSpT]], type[NameSpT]]:
    """Decorator for registering custom functionality with an :class:`~anndata.AnnData` object.

    This decorator allows you to extend AnnData objects with custom methods and properties
    organized under a namespace. The namespace becomes accessible as an attribute on AnnData
    instances, providing a clean way to you to add domain-specific functionality without modifying
    the AnnData class itself, or extending the class with additional methods as you see fit in your workflow.

    Parameters
    ----------
    name
        Name under which the accessor should be registered. This will be the attribute name
        used to access your namespace's functionality on AnnData objects (e.g., `adata.{name}`).
        Cannot conflict with existing AnnData attributes like `obs`, `var`, `X`, etc. The list of reserved
        attributes includes everything outputted by `dir(AnnData)`.

    Returns
    -------
    A decorator that registers the decorated class as a custom namespace.

    Notes
    -----
    Implementation requirements:

    1. The decorated class must have an `__init__` method that accepts exactly one parameter
       (besides `self`) named `adata` and annotated with type :class:`~anndata.AnnData`.
    2. The namespace will be initialized with the AnnData object on first access and then
       cached on the instance.
    3. If the namespace name conflicts with an existing namespace, a warning is issued.
    4. If the namespace name conflicts with a built-in AnnData attribute, an AttributeError is raised.

    Examples
    --------
    Simple transformation namespace with two methods:

    >>> import anndata as ad
    >>> import numpy as np
    >>>
    >>> @ad.register_anndata_namespace("transform")
    ... class TransformX:
    ...     def __init__(self, adata: ad.AnnData):
    ...         self._adata = adata
    ...
    ...     def log1p(
    ...         self, layer: str = None, inplace: bool = False
    ...     ) -> ad.AnnData | None:
    ...         '''Log1p transform the data.'''
    ...         data = self._adata.layers[layer] if layer else self._adata.X
    ...         log1p_data = np.log1p(data)
    ...
    ...         if layer:
    ...             layer_name = f"{layer}_log1p" if not inplace else layer
    ...         else:
    ...             layer_name = "log1p"
    ...
    ...         self._adata.layers[layer_name] = log1p_data
    ...
    ...         if not inplace:
    ...             return self._adata
    ...
    ...     def arcsinh(
    ...         self, layer: str = None, scale: float = 1.0, inplace: bool = False
    ...     ) -> ad.AnnData | None:
    ...         '''Arcsinh transform the data with optional scaling.'''
    ...         data = self._adata.layers[layer] if layer else self._adata.X
    ...         asinh_data = np.arcsinh(data / scale)
    ...
    ...         if layer:
    ...             layer_name = f"{layer}_arcsinh" if not inplace else layer
    ...         else:
    ...             layer_name = "arcsinh"
    ...
    ...         self._adata.layers[layer_name] = asinh_data
    ...
    ...         if not inplace:
    ...             return self._adata
    >>>
    >>> # Create an AnnData object
    >>> rng = np.random.default_rng(42)
    >>> adata = ad.AnnData(X=rng.poisson(1, size=(100, 2000)))
    >>>
    >>> # Use the registered namespace
    >>> adata.transform.log1p()  # Transforms X and returns the AnnData object
    AnnData object with n_obs × n_vars = 100 × 2000
        layers: 'log1p'
    >>> adata.transform.arcsinh()  # Transforms X and returns the AnnData object
    AnnData object with n_obs × n_vars = 100 × 2000
        layers: 'log1p', 'arcsinh'
    """
    return _create_namespace(name, AnnData)
